@@ -5,6 +5,7 @@
 #include <stdint.h>
 #include <cairo.h>
 #include <pango/pangocairo.h>
+#include "stringop.h"
 
 /**
  * Utility function which escape characters a & < > ' ".
@@ -16,9 +17,9 @@ size_t escape_markup_text(const char *src, char *dest);
 PangoLayout *get_pango_layout(cairo_t *cairo, const PangoFontDescription *desc,
 		const char *text, double scale, bool markup);
 void get_text_size(cairo_t *cairo, const PangoFontDescription *desc, int *width, int *height,
-		int *baseline, double scale, bool markup, const char *fmt, ...);
+		int *baseline, double scale, bool markup, const char *fmt, ...) _SWAY_ATTRIB_PRINTF(8, 9);
 void get_text_metrics(const PangoFontDescription *desc, int *height, int *baseline);
 void render_text(cairo_t *cairo, PangoFontDescription *desc,
-		double scale, bool markup, const char *fmt, ...);
+		double scale, bool markup, const char *fmt, ...) _SWAY_ATTRIB_PRINTF(5, 6);
 
 #endif
