@@ -1,4 +1,4 @@
-#include <wlr/types/wlr_surface.h>
+#include <wlr/types/wlr_compositor.h>
 
 struct sway_container;
 struct sway_view;
