@@ -15,4 +15,10 @@ struct sway_surface {
 	struct wl_event_source *frame_done_timer;
 };
 
+void surface_update_outputs(struct wlr_surface *surface);
+void surface_enter_output(struct wlr_surface *surface,
+	struct sway_output *output);
+void surface_leave_output(struct wlr_surface *surface,
+	struct sway_output *output);
+
 #endif
