@@ -1,8 +1,8 @@
 #ifndef _SWAY_VIEW_H
 #define _SWAY_VIEW_H
 #include <wayland-server-core.h>
-#include <wlr/types/wlr_surface.h>
-#include "config.h"
+#include <wlr/types/wlr_compositor.h>
+#include "sway/config.h"
 #if HAVE_XWAYLAND
 #include <wlr/xwayland.h>
 #endif
@@ -74,6 +74,7 @@ struct sway_view {
 	struct sway_xdg_decoration *xdg_decoration;
 
 	pid_t pid;
+	struct launcher_ctx *ctx;
 
 	// The size the view would want to be if it weren't tiled.
 	// Used when changing a view from tiled to floating.
@@ -108,7 +109,7 @@ struct sway_view {
 	list_t *executed_criteria; // struct criteria *
 
 	union {
-		struct wlr_xdg_surface *wlr_xdg_surface;
+		struct wlr_xdg_toplevel *wlr_xdg_toplevel;
 #if HAVE_XWAYLAND
 		struct wlr_xwayland_surface *wlr_xwayland_surface;
 #endif
@@ -131,6 +132,7 @@ struct sway_xdg_shell_view {
 	struct wl_listener commit;
 	struct wl_listener request_move;
 	struct wl_listener request_resize;
+	struct wl_listener request_maximize;
 	struct wl_listener request_fullscreen;
 	struct wl_listener set_title;
 	struct wl_listener set_app_id;
@@ -154,9 +156,12 @@ struct sway_xwayland_view {
 	struct wl_listener set_title;
 	struct wl_listener set_class;
 	struct wl_listener set_role;
+	struct wl_listener set_startup_id;
 	struct wl_listener set_window_type;
 	struct wl_listener set_hints;
 	struct wl_listener set_decorations;
+	struct wl_listener associate;
+	struct wl_listener dissociate;
 	struct wl_listener map;
 	struct wl_listener unmap;
 	struct wl_listener destroy;
@@ -169,10 +174,13 @@ struct sway_xwayland_unmanaged {
 
 	int lx, ly;
 
+	struct wl_listener request_activate;
 	struct wl_listener request_configure;
 	struct wl_listener request_fullscreen;
 	struct wl_listener commit;
 	struct wl_listener set_geometry;
+	struct wl_listener associate;
+	struct wl_listener dissociate;
 	struct wl_listener map;
 	struct wl_listener unmap;
 	struct wl_listener destroy;
@@ -216,7 +224,7 @@ struct sway_subsurface {
 struct sway_xdg_popup {
 	struct sway_view_child child;
 
-	struct wlr_xdg_surface *wlr_xdg_surface;
+	struct wlr_xdg_popup *wlr_xdg_popup;
 
 	struct wl_listener new_popup;
 	struct wl_listener destroy;
@@ -267,7 +275,12 @@ void view_set_activated(struct sway_view *view, bool activated);
 /**
  * Called when the view requests to be focused.
  */
-void view_request_activate(struct sway_view *view);
+void view_request_activate(struct sway_view *view, struct sway_seat *seat);
+
+/*
+ * Called when the view requests urgent state
+ */
+void view_request_urgent(struct sway_view *view);
 
 /**
  * If possible, instructs the client to change their decoration mode.
@@ -370,4 +383,6 @@ void view_save_buffer(struct sway_view *view);
 
 bool view_is_transient_for(struct sway_view *child, struct sway_view *ancestor);
 
+void view_assign_ctx(struct sway_view *view, struct launcher_ctx *ctx);
+
 #endif
