@@ -3,7 +3,8 @@
 #include <stdio.h>
 #include <stdbool.h>
 #include <strings.h>
-#include <pcre.h>
+#define PCRE2_CODE_UNIT_WIDTH 8
+#include <pcre2.h>
 #include "sway/criteria.h"
 #include "sway/tree/container.h"
 #include "sway/config.h"
@@ -18,6 +19,7 @@
 bool criteria_is_empty(struct criteria *criteria) {
 	return !criteria->title
 		&& !criteria->shell
+		&& !criteria->all
 		&& !criteria->app_id
 		&& !criteria->con_mark
 		&& !criteria->con_id
@@ -40,17 +42,19 @@ bool criteria_is_empty(struct criteria *criteria) {
 char *error = NULL;
 
 // Returns error string on failure or NULL otherwise.
-static bool generate_regex(pcre **regex, char *value) {
-	const char *reg_err;
-	int offset;
-
-	*regex = pcre_compile(value, PCRE_UTF8 | PCRE_UCP, &reg_err, &offset, NULL);
+static bool generate_regex(pcre2_code **regex, char *value) {
+	int errorcode;
+	PCRE2_SIZE offset;
 
+	*regex = pcre2_compile((PCRE2_SPTR)value, PCRE2_ZERO_TERMINATED, PCRE2_UTF | PCRE2_UCP, &errorcode, &offset, NULL);
 	if (!*regex) {
+		PCRE2_UCHAR buffer[256];
+		pcre2_get_error_message(errorcode, buffer, sizeof(buffer));
+
 		const char *fmt = "Regex compilation for '%s' failed: %s";
-		int len = strlen(fmt) + strlen(value) + strlen(reg_err) - 3;
+		int len = strlen(fmt) + strlen(value) + strlen((char*) buffer) - 3;
 		error = malloc(len);
-		snprintf(error, len, fmt, value, reg_err);
+		snprintf(error, len, fmt, value, buffer);
 		return false;
 	}
 
@@ -66,7 +70,7 @@ static bool pattern_create(struct pattern **pattern, char *value) {
 	if (strcmp(value, "__focused__") == 0) {
 		(*pattern)->match_type = PATTERN_FOCUSED;
 	} else {
-		(*pattern)->match_type = PATTERN_PCRE;
+		(*pattern)->match_type = PATTERN_PCRE2;
 		if (!generate_regex(&(*pattern)->regex, value)) {
 			return false;
 		};
@@ -77,7 +81,7 @@ static bool pattern_create(struct pattern **pattern, char *value) {
 static void pattern_destroy(struct pattern *pattern) {
 	if (pattern) {
 		if (pattern->regex) {
-			pcre_free(pattern->regex);
+			pcre2_code_free(pattern->regex);
 		}
 		free(pattern);
 	}
@@ -93,14 +97,18 @@ void criteria_destroy(struct criteria *criteria) {
 	pattern_destroy(criteria->window_role);
 #endif
 	pattern_destroy(criteria->con_mark);
-	free(criteria->workspace);
+	pattern_destroy(criteria->workspace);
+	free(criteria->target);
 	free(criteria->cmdlist);
 	free(criteria->raw);
 	free(criteria);
 }
 
-static int regex_cmp(const char *item, const pcre *regex) {
-	return pcre_exec(regex, NULL, item, strlen(item), 0, 0, NULL, 0);
+static int regex_cmp(const char *item, const pcre2_code *regex) {
+	pcre2_match_data *match_data = pcre2_match_data_create_from_pattern(regex, NULL);
+	int result = pcre2_match(regex, (PCRE2_SPTR)item, strlen(item), 0, 0, match_data, NULL);
+	pcre2_match_data_free(match_data);
+	return result;
 }
 
 #if HAVE_XWAYLAND
@@ -155,7 +163,7 @@ static bool criteria_matches_container(struct criteria *criteria,
 		bool exists = false;
 		struct sway_container *con = container;
 		for (int i = 0; i < con->marks->length; ++i) {
-			if (regex_cmp(con->marks->items[i], criteria->con_mark->regex) == 0) {
+			if (regex_cmp(con->marks->items[i], criteria->con_mark->regex) >= 0) {
 				exists = true;
 				break;
 			}
@@ -183,7 +191,7 @@ static bool criteria_matches_view(struct criteria *criteria,
 	if (criteria->title) {
 		const char *title = view_get_title(view);
 		if (!title) {
-			return false;
+			title = "";
 		}
 
 		switch (criteria->title->match_type) {
@@ -192,8 +200,8 @@ static bool criteria_matches_view(struct criteria *criteria,
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(title, criteria->title->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(title, criteria->title->regex) < 0) {
 				return false;
 			}
 			break;
@@ -203,7 +211,7 @@ static bool criteria_matches_view(struct criteria *criteria,
 	if (criteria->shell) {
 		const char *shell = view_get_shell(view);
 		if (!shell) {
-			return false;
+			shell = "";
 		}
 
 		switch (criteria->shell->match_type) {
@@ -212,8 +220,8 @@ static bool criteria_matches_view(struct criteria *criteria,
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(shell, criteria->shell->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(shell, criteria->shell->regex) < 0) {
 				return false;
 			}
 			break;
@@ -223,7 +231,7 @@ static bool criteria_matches_view(struct criteria *criteria,
 	if (criteria->app_id) {
 		const char *app_id = view_get_app_id(view);
 		if (!app_id) {
-			return false;
+			app_id = "";
 		}
 
 		switch (criteria->app_id->match_type) {
@@ -232,8 +240,8 @@ static bool criteria_matches_view(struct criteria *criteria,
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(app_id, criteria->app_id->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(app_id, criteria->app_id->regex) < 0) {
 				return false;
 			}
 			break;
@@ -255,7 +263,7 @@ static bool criteria_matches_view(struct criteria *criteria,
 	if (criteria->class) {
 		const char *class = view_get_class(view);
 		if (!class) {
-			return false;
+			class = "";
 		}
 
 		switch (criteria->class->match_type) {
@@ -264,8 +272,8 @@ static bool criteria_matches_view(struct criteria *criteria,
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(class, criteria->class->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(class, criteria->class->regex) < 0) {
 				return false;
 			}
 			break;
@@ -275,17 +283,17 @@ static bool criteria_matches_view(struct criteria *criteria,
 	if (criteria->instance) {
 		const char *instance = view_get_instance(view);
 		if (!instance) {
-			return false;
+			instance = "";
 		}
 
 		switch (criteria->instance->match_type) {
 		case PATTERN_FOCUSED:
-			if (focused && strcmp(instance, view_get_instance(focused))) {
+			if (focused && lenient_strcmp(instance, view_get_instance(focused))) {
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(instance, criteria->instance->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(instance, criteria->instance->regex) < 0) {
 				return false;
 			}
 			break;
@@ -295,17 +303,17 @@ static bool criteria_matches_view(struct criteria *criteria,
 	if (criteria->window_role) {
 		const char *window_role = view_get_window_role(view);
 		if (!window_role) {
-			return false;
+			window_role = "";
 		}
 
 		switch (criteria->window_role->match_type) {
 		case PATTERN_FOCUSED:
-			if (focused && strcmp(window_role, view_get_window_role(focused))) {
+			if (focused && lenient_strcmp(window_role, view_get_window_role(focused))) {
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(window_role, criteria->window_role->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(window_role, criteria->window_role->regex) < 0) {
 				return false;
 			}
 			break;
@@ -363,8 +371,8 @@ static bool criteria_matches_view(struct criteria *criteria,
 				return false;
 			}
 			break;
-		case PATTERN_PCRE:
-			if (regex_cmp(ws->name, criteria->workspace->regex) != 0) {
+		case PATTERN_PCRE2:
+			if (regex_cmp(ws->name, criteria->workspace->regex) < 0) {
 				return false;
 			}
 			break;
@@ -449,6 +457,7 @@ static enum atom_name parse_window_type(const char *type) {
 #endif
 
 enum criteria_token {
+	T_ALL,
 	T_APP_ID,
 	T_CON_ID,
 	T_CON_MARK,
@@ -471,7 +480,9 @@ enum criteria_token {
 };
 
 static enum criteria_token token_from_name(char *name) {
-	if (strcmp(name, "app_id") == 0) {
+	if (strcmp(name, "all") == 0) {
+		return T_ALL;
+	} else if (strcmp(name, "app_id") == 0) {
 		return T_APP_ID;
 	} else if (strcmp(name, "con_id") == 0) {
 		return T_CON_ID;
@@ -517,8 +528,8 @@ static bool parse_token(struct criteria *criteria, char *name, char *value) {
 		return false;
 	}
 
-	// Require value, unless token is floating or tiled
-	if (!value && token != T_FLOATING && token != T_TILING) {
+	// Require value, unless token is all, floating or tiled
+	if (!value && token != T_ALL && token != T_FLOATING && token != T_TILING) {
 		const char *fmt = "Token '%s' requires a value";
 		int len = strlen(fmt) + strlen(name) - 1;
 		error = malloc(len);
@@ -528,6 +539,9 @@ static bool parse_token(struct criteria *criteria, char *name, char *value) {
 
 	char *endptr = NULL;
 	switch (token) {
+	case T_ALL:
+		criteria->all = true;
+		break;
 	case T_TITLE:
 		pattern_create(&criteria->title, value);
 		break;
@@ -676,7 +690,7 @@ struct criteria *criteria_parse(char *raw, char **error_arg) {
 		}
 		name = calloc(head - namestart + 1, 1);
 		if (head != namestart) {
-			strncpy(name, namestart, head - namestart);
+			memcpy(name, namestart, head - namestart);
 		}
 		// Parse token value
 		skip_spaces(&head);
@@ -703,7 +717,7 @@ struct criteria *criteria_parse(char *raw, char **error_arg) {
 				}
 			}
 			value = calloc(head - valuestart + 1, 1);
-			strncpy(value, valuestart, head - valuestart);
+			memcpy(value, valuestart, head - valuestart);
 			if (in_quotes) {
 				++head;
 				in_quotes = false;
@@ -734,7 +748,7 @@ struct criteria *criteria_parse(char *raw, char **error_arg) {
 	++head;
 	int len = head - raw;
 	criteria->raw = calloc(len + 1, 1);
-	strncpy(criteria->raw, raw, len);
+	memcpy(criteria->raw, raw, len);
 	return criteria;
 
 cleanup:
