@@ -3,8 +3,8 @@
 #include <string.h>
 #include <wayland-server-core.h>
 #include <wlr/types/wlr_layer_shell_v1.h>
-#include <wlr/types/wlr_output_damage.h>
 #include <wlr/types/wlr_output.h>
+#include <wlr/types/wlr_subcompositor.h>
 #include "log.h"
 #include "sway/desktop/transaction.h"
 #include "sway/input/cursor.h"
@@ -13,9 +13,43 @@
 #include "sway/layers.h"
 #include "sway/output.h"
 #include "sway/server.h"
+#include "sway/surface.h"
 #include "sway/tree/arrange.h"
 #include "sway/tree/workspace.h"
 
+struct wlr_layer_surface_v1 *toplevel_layer_surface_from_surface(
+		struct wlr_surface *surface) {
+	struct wlr_layer_surface_v1 *layer;
+	do {
+		if (!surface) {
+			return NULL;
+		}
+		// Topmost layer surface
+		if ((layer = wlr_layer_surface_v1_try_from_wlr_surface(surface))) {
+			return layer;
+		}
+		// Layer subsurface
+		if (wlr_subsurface_try_from_wlr_surface(surface)) {
+			surface = wlr_surface_get_root_surface(surface);
+			continue;
+		}
+
+		// Layer surface popup
+		struct wlr_xdg_surface *xdg_surface = NULL;
+		if ((xdg_surface = wlr_xdg_surface_try_from_wlr_surface(surface)) &&
+				xdg_surface->role == WLR_XDG_SURFACE_ROLE_POPUP && xdg_surface->popup != NULL) {
+			if (!xdg_surface->popup->parent) {
+				return NULL;
+			}
+			surface = wlr_surface_get_root_surface(xdg_surface->popup->parent);
+			continue;
+		}
+
+		// Return early if the surface is not a layer/xdg_popup/sub surface
+		return NULL;
+	} while (true);
+}
+
 static void apply_exclusive(struct wlr_box *usable_area,
 		uint32_t anchor, int32_t exclusive,
 		int32_t margin_top, int32_t margin_right,
@@ -217,8 +251,9 @@ void arrange_layers(struct sway_output *output) {
 	for (size_t i = 0; i < nlayers; ++i) {
 		wl_list_for_each_reverse(layer,
 				&output->layers[layers_above_shell[i]], link) {
-			if (layer->layer_surface->current.keyboard_interactive &&
-					layer->layer_surface->mapped) {
+			if (layer->layer_surface->current.keyboard_interactive
+					== ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE &&
+					layer->layer_surface->surface->mapped) {
 				topmost = layer;
 				break;
 			}
@@ -230,10 +265,12 @@ void arrange_layers(struct sway_output *output) {
 
 	struct sway_seat *seat;
 	wl_list_for_each(seat, &server.input->seats, link) {
+		seat->has_exclusive_layer = false;
 		if (topmost != NULL) {
 			seat_set_focus_layer(seat, topmost->layer_surface);
 		} else if (seat->focused_layer &&
-				!seat->focused_layer->current.keyboard_interactive) {
+				seat->focused_layer->current.keyboard_interactive
+					!= ZWLR_LAYER_SURFACE_V1_KEYBOARD_INTERACTIVITY_EXCLUSIVE) {
 			seat_set_focus_layer(seat, NULL);
 		}
 	}
@@ -252,7 +289,7 @@ static struct sway_layer_surface *find_mapped_layer_by_client(
 				&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY], link) {
 			struct wl_resource *resource = lsurface->layer_surface->resource;
 			if (wl_resource_get_client(resource) == client
-					&& lsurface->layer_surface->mapped) {
+					&& lsurface->layer_surface->surface->mapped) {
 				return lsurface;
 			}
 		}
@@ -270,10 +307,6 @@ static void handle_output_destroy(struct wl_listener *listener, void *data) {
 		wl_resource_get_client(sway_layer->layer_surface->resource);
 	bool set_focus = seat->exclusive_client == client;
 
-	wl_list_remove(&sway_layer->output_destroy.link);
-	wl_list_remove(&sway_layer->link);
-	wl_list_init(&sway_layer->link);
-
 	if (set_focus) {
 		struct sway_layer_surface *layer =
 			find_mapped_layer_by_client(client, sway_layer->layer_surface->output);
@@ -282,7 +315,6 @@ static void handle_output_destroy(struct wl_listener *listener, void *data) {
 		}
 	}
 
-	sway_layer->layer_surface->output = NULL;
 	wlr_layer_surface_v1_destroy(sway_layer->layer_surface);
 }
 
@@ -291,17 +323,14 @@ static void handle_surface_commit(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, layer, surface_commit);
 	struct wlr_layer_surface_v1 *layer_surface = layer->layer_surface;
 	struct wlr_output *wlr_output = layer_surface->output;
-	if (wlr_output == NULL) {
-		return;
-	}
-
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
 	struct sway_output *output = wlr_output->data;
 	struct wlr_box old_extent = layer->extent;
 
 	bool layer_changed = false;
 	if (layer_surface->current.committed != 0
-			|| layer->mapped != layer_surface->mapped) {
-		layer->mapped = layer_surface->mapped;
+			|| layer->mapped != layer_surface->surface->mapped) {
+		layer->mapped = layer_surface->surface->mapped;
 		layer_changed = layer->layer != layer_surface->current.layer;
 		if (layer_changed) {
 			wl_list_remove(&layer->link);
@@ -319,6 +348,8 @@ static void handle_surface_commit(struct wl_listener *listener, void *data) {
 	bool extent_changed =
 		memcmp(&old_extent, &layer->extent, sizeof(struct wlr_box)) != 0;
 	if (extent_changed || layer_changed) {
+		old_extent.x += output->lx;
+		old_extent.y += output->ly;
 		output_damage_box(output, &old_extent);
 		output_damage_surface(output, layer->geo.x, layer->geo.y,
 			layer_surface->surface, true);
@@ -341,13 +372,8 @@ static void unmap(struct sway_layer_surface *sway_layer) {
 	cursor_rebase_all();
 
 	struct wlr_output *wlr_output = sway_layer->layer_surface->output;
-	if (wlr_output == NULL) {
-		return;
-	}
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
 	struct sway_output *output = wlr_output->data;
-	if (output == NULL) {
-		return;
-	}
 	output_damage_surface(output, sway_layer->geo.x, sway_layer->geo.y,
 		sway_layer->layer_surface->surface, true);
 }
@@ -359,7 +385,7 @@ static void handle_destroy(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, sway_layer, destroy);
 	sway_log(SWAY_DEBUG, "Layer surface destroyed (%s)",
 		sway_layer->layer_surface->namespace);
-	if (sway_layer->layer_surface->mapped) {
+	if (sway_layer->layer_surface->surface->mapped) {
 		unmap(sway_layer);
 	}
 
@@ -375,26 +401,26 @@ static void handle_destroy(struct wl_listener *listener, void *data) {
 	wl_list_remove(&sway_layer->surface_commit.link);
 	wl_list_remove(&sway_layer->new_popup.link);
 	wl_list_remove(&sway_layer->new_subsurface.link);
-	if (sway_layer->layer_surface->output != NULL) {
-		struct sway_output *output = sway_layer->layer_surface->output->data;
-		if (output != NULL) {
-			arrange_layers(output);
-			transaction_commit_dirty();
-		}
-		wl_list_remove(&sway_layer->output_destroy.link);
-		sway_layer->layer_surface->output = NULL;
-	}
+
+	struct wlr_output *wlr_output = sway_layer->layer_surface->output;
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
+	struct sway_output *output = wlr_output->data;
+	arrange_layers(output);
+	transaction_commit_dirty();
+	wl_list_remove(&sway_layer->output_destroy.link);
+	sway_layer->layer_surface->output = NULL;
+
 	free(sway_layer);
 }
 
 static void handle_map(struct wl_listener *listener, void *data) {
 	struct sway_layer_surface *sway_layer = wl_container_of(listener,
 			sway_layer, map);
-	struct sway_output *output = sway_layer->layer_surface->output->data;
+	struct wlr_output *wlr_output = sway_layer->layer_surface->output;
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
+	struct sway_output *output = wlr_output->data;
 	output_damage_surface(output, sway_layer->geo.x, sway_layer->geo.y,
 		sway_layer->layer_surface->surface, true);
-	wlr_surface_send_enter(sway_layer->layer_surface->surface,
-		sway_layer->layer_surface->output);
 	cursor_rebase_all();
 }
 
@@ -408,9 +434,7 @@ static void subsurface_damage(struct sway_layer_subsurface *subsurface,
 		bool whole) {
 	struct sway_layer_surface *layer = subsurface->layer_surface;
 	struct wlr_output *wlr_output = layer->layer_surface->output;
-	if (!wlr_output) {
-		return;
-	}
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
 	struct sway_output *output = wlr_output->data;
 	int ox = subsurface->wlr_subsurface->current.x + layer->geo.x;
 	int oy = subsurface->wlr_subsurface->current.y + layer->geo.y;
@@ -466,9 +490,9 @@ static struct sway_layer_subsurface *create_subsurface(
 	wl_list_insert(&layer_surface->subsurfaces, &subsurface->link);
 
 	subsurface->map.notify = subsurface_handle_map;
-	wl_signal_add(&wlr_subsurface->events.map, &subsurface->map);
+	wl_signal_add(&wlr_subsurface->surface->events.map, &subsurface->map);
 	subsurface->unmap.notify = subsurface_handle_unmap;
-	wl_signal_add(&wlr_subsurface->events.unmap, &subsurface->unmap);
+	wl_signal_add(&wlr_subsurface->surface->events.unmap, &subsurface->unmap);
 	subsurface->destroy.notify = subsurface_handle_destroy;
 	wl_signal_add(&wlr_subsurface->events.destroy, &subsurface->destroy);
 	subsurface->commit.notify = subsurface_handle_commit;
@@ -496,15 +520,15 @@ static struct sway_layer_surface *popup_get_layer(
 static void popup_damage(struct sway_layer_popup *layer_popup, bool whole) {
 	struct wlr_xdg_popup *popup = layer_popup->wlr_popup;
 	struct wlr_surface *surface = popup->base->surface;
-	int popup_sx = popup->geometry.x - popup->base->current.geometry.x;
-	int popup_sy = popup->geometry.y - popup->base->current.geometry.y;
+	int popup_sx = popup->current.geometry.x - popup->base->current.geometry.x;
+	int popup_sy = popup->current.geometry.y - popup->base->current.geometry.y;
 	int ox = popup_sx, oy = popup_sy;
 	struct sway_layer_surface *layer;
 	while (true) {
 		if (layer_popup->parent_type == LAYER_PARENT_POPUP) {
 			layer_popup = layer_popup->parent_popup;
-			ox += layer_popup->wlr_popup->geometry.x;
-			oy += layer_popup->wlr_popup->geometry.y;
+			ox += layer_popup->wlr_popup->current.geometry.x;
+			oy += layer_popup->wlr_popup->current.geometry.y;
 		} else {
 			layer = layer_popup->parent_layer;
 			ox += layer->geo.x;
@@ -513,6 +537,7 @@ static void popup_damage(struct sway_layer_popup *layer_popup, bool whole) {
 		}
 	}
 	struct wlr_output *wlr_output = layer->layer_surface->output;
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
 	struct sway_output *output = wlr_output->data;
 	output_damage_surface(output, ox, oy, surface, whole);
 }
@@ -521,7 +546,8 @@ static void popup_handle_map(struct wl_listener *listener, void *data) {
 	struct sway_layer_popup *popup = wl_container_of(listener, popup, map);
 	struct sway_layer_surface *layer = popup_get_layer(popup);
 	struct wlr_output *wlr_output = layer->layer_surface->output;
-	wlr_surface_send_enter(popup->wlr_popup->base->surface, wlr_output);
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
+	surface_enter_output(popup->wlr_popup->base->surface, wlr_output->data);
 	popup_damage(popup, true);
 }
 
@@ -550,7 +576,9 @@ static void popup_unconstrain(struct sway_layer_popup *popup) {
 	struct sway_layer_surface *layer = popup_get_layer(popup);
 	struct wlr_xdg_popup *wlr_popup = popup->wlr_popup;
 
-	struct sway_output *output = layer->layer_surface->output->data;
+	struct wlr_output *wlr_output = layer->layer_surface->output;
+	sway_assert(wlr_output, "wlr_layer_surface_v1 has null output");
+	struct sway_output *output = wlr_output->data;
 
 	// the output box expressed in the coordinate system of the toplevel parent
 	// of the popup
@@ -579,9 +607,9 @@ static struct sway_layer_popup *create_popup(struct wlr_xdg_popup *wlr_popup,
 	popup->parent_layer = parent;
 
 	popup->map.notify = popup_handle_map;
-	wl_signal_add(&wlr_popup->base->events.map, &popup->map);
+	wl_signal_add(&wlr_popup->base->surface->events.map, &popup->map);
 	popup->unmap.notify = popup_handle_unmap;
-	wl_signal_add(&wlr_popup->base->events.unmap, &popup->unmap);
+	wl_signal_add(&wlr_popup->base->surface->events.unmap, &popup->unmap);
 	popup->destroy.notify = popup_handle_destroy;
 	wl_signal_add(&wlr_popup->base->events.destroy, &popup->destroy);
 	popup->commit.notify = popup_handle_commit;
@@ -642,6 +670,10 @@ void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 				sway_log(SWAY_ERROR,
 						"no output to auto-assign layer surface '%s' to",
 						layer_surface->namespace);
+				// Note that layer_surface->output can be NULL
+				// here, but none of our destroy callbacks are
+				// registered yet so we don't have to make them
+				// handle that case.
 				wlr_layer_surface_v1_destroy(layer_surface);
 				return;
 			}
@@ -665,9 +697,9 @@ void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 	sway_layer->destroy.notify = handle_destroy;
 	wl_signal_add(&layer_surface->events.destroy, &sway_layer->destroy);
 	sway_layer->map.notify = handle_map;
-	wl_signal_add(&layer_surface->events.map, &sway_layer->map);
+	wl_signal_add(&layer_surface->surface->events.map, &sway_layer->map);
 	sway_layer->unmap.notify = handle_unmap;
-	wl_signal_add(&layer_surface->events.unmap, &sway_layer->unmap);
+	wl_signal_add(&layer_surface->surface->events.unmap, &sway_layer->unmap);
 	sway_layer->new_popup.notify = handle_new_popup;
 	wl_signal_add(&layer_surface->events.new_popup, &sway_layer->new_popup);
 	sway_layer->new_subsurface.notify = handle_new_subsurface;
@@ -684,6 +716,8 @@ void handle_layer_shell_surface(struct wl_listener *listener, void *data) {
 	wl_list_insert(&output->layers[layer_surface->pending.layer],
 			&sway_layer->link);
 
+	surface_enter_output(layer_surface->surface, output);
+
 	// Temporarily set the layer's current state to pending
 	// So that we can easily arrange it
 	struct wlr_layer_surface_v1_state old_state = layer_surface->current;
