@@ -5,7 +5,9 @@
 #include <wayland-server-core.h>
 #include <wlr/types/wlr_output_layout.h>
 #include <wlr/types/wlr_output.h>
+#include <wlr/types/wlr_xdg_activation_v1.h>
 #include <wlr/xwayland.h>
+#include <xcb/xcb_icccm.h>
 #include "log.h"
 #include "sway/desktop.h"
 #include "sway/desktop/transaction.h"
@@ -15,6 +17,7 @@
 #include "sway/output.h"
 #include "sway/tree/arrange.h"
 #include "sway/tree/container.h"
+#include "sway/server.h"
 #include "sway/tree/view.h"
 #include "sway/tree/workspace.h"
 
@@ -121,18 +124,53 @@ static void unmanaged_handle_unmap(struct wl_listener *listener, void *data) {
 	}
 }
 
+static void unmanaged_handle_request_activate(struct wl_listener *listener, void *data) {
+	struct sway_xwayland_unmanaged *surface =
+		wl_container_of(listener, surface, request_activate);
+	struct wlr_xwayland_surface *xsurface = surface->wlr_xwayland_surface;
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
+		return;
+	}
+	struct sway_seat *seat = input_manager_current_seat();
+	struct sway_container *focus = seat_get_focused_container(seat);
+	if (focus && focus->view && focus->view->pid != xsurface->pid) {
+		return;
+	}
+
+	seat_set_focus_surface(seat, xsurface->surface, false);
+}
+
+static void unmanaged_handle_associate(struct wl_listener *listener, void *data) {
+	struct sway_xwayland_unmanaged *surface =
+		wl_container_of(listener, surface, associate);
+	struct wlr_xwayland_surface *xsurface = surface->wlr_xwayland_surface;
+	wl_signal_add(&xsurface->surface->events.map, &surface->map);
+	surface->map.notify = unmanaged_handle_map;
+	wl_signal_add(&xsurface->surface->events.unmap, &surface->unmap);
+	surface->unmap.notify = unmanaged_handle_unmap;
+}
+
+static void unmanaged_handle_dissociate(struct wl_listener *listener, void *data) {
+	struct sway_xwayland_unmanaged *surface =
+		wl_container_of(listener, surface, dissociate);
+	wl_list_remove(&surface->map.link);
+	wl_list_remove(&surface->unmap.link);
+}
+
 static void unmanaged_handle_destroy(struct wl_listener *listener, void *data) {
 	struct sway_xwayland_unmanaged *surface =
 		wl_container_of(listener, surface, destroy);
 	wl_list_remove(&surface->request_configure.link);
-	wl_list_remove(&surface->map.link);
-	wl_list_remove(&surface->unmap.link);
+	wl_list_remove(&surface->associate.link);
+	wl_list_remove(&surface->dissociate.link);
 	wl_list_remove(&surface->destroy.link);
 	wl_list_remove(&surface->override_redirect.link);
+	wl_list_remove(&surface->request_activate.link);
 	free(surface);
 }
 
 static void handle_map(struct wl_listener *listener, void *data);
+static void handle_associate(struct wl_listener *listener, void *data);
 
 struct sway_xwayland_view *create_xwayland_view(struct wlr_xwayland_surface *xsurface);
 
@@ -141,14 +179,22 @@ static void unmanaged_handle_override_redirect(struct wl_listener *listener, voi
 		wl_container_of(listener, surface, override_redirect);
 	struct wlr_xwayland_surface *xsurface = surface->wlr_xwayland_surface;
 
-	bool mapped = xsurface->mapped;
+	bool associated = xsurface->surface != NULL;
+	bool mapped = associated && xsurface->surface->mapped;
 	if (mapped) {
 		unmanaged_handle_unmap(&surface->unmap, NULL);
 	}
+	if (associated) {
+		unmanaged_handle_dissociate(&surface->dissociate, NULL);
+	}
 
 	unmanaged_handle_destroy(&surface->destroy, NULL);
 	xsurface->data = NULL;
+
 	struct sway_xwayland_view *xwayland_view = create_xwayland_view(xsurface);
+	if (associated) {
+		handle_associate(&xwayland_view->associate, NULL);
+	}
 	if (mapped) {
 		handle_map(&xwayland_view->map, xsurface);
 	}
@@ -168,14 +214,16 @@ static struct sway_xwayland_unmanaged *create_unmanaged(
 	wl_signal_add(&xsurface->events.request_configure,
 		&surface->request_configure);
 	surface->request_configure.notify = unmanaged_handle_request_configure;
-	wl_signal_add(&xsurface->events.map, &surface->map);
-	surface->map.notify = unmanaged_handle_map;
-	wl_signal_add(&xsurface->events.unmap, &surface->unmap);
-	surface->unmap.notify = unmanaged_handle_unmap;
+	wl_signal_add(&xsurface->events.associate, &surface->associate);
+	surface->associate.notify = unmanaged_handle_associate;
+	wl_signal_add(&xsurface->events.dissociate, &surface->dissociate);
+	surface->dissociate.notify = unmanaged_handle_dissociate;
 	wl_signal_add(&xsurface->events.destroy, &surface->destroy);
 	surface->destroy.notify = unmanaged_handle_destroy;
 	wl_signal_add(&xsurface->events.set_override_redirect, &surface->override_redirect);
 	surface->override_redirect.notify = unmanaged_handle_override_redirect;
+	wl_signal_add(&xsurface->events.request_activate, &surface->request_activate);
+	surface->request_activate.notify = unmanaged_handle_request_activate;
 
 	return surface;
 }
@@ -294,7 +342,7 @@ static bool wants_floating(struct sway_view *view) {
 		}
 	}
 
-	struct wlr_xwayland_surface_size_hints *size_hints = surface->size_hints;
+	xcb_size_hints_t *size_hints = surface->size_hints;
 	if (size_hints != NULL &&
 			size_hints->min_width > 0 && size_hints->min_height > 0 &&
 			(size_hints->max_width == size_hints->min_width ||
@@ -348,7 +396,7 @@ static void destroy(struct sway_view *view) {
 static void get_constraints(struct sway_view *view, double *min_width,
 		double *max_width, double *min_height, double *max_height) {
 	struct wlr_xwayland_surface *surface = view->wlr_xwayland_surface;
-	struct wlr_xwayland_surface_size_hints *size_hints = surface->size_hints;
+	xcb_size_hints_t *size_hints = surface->size_hints;
 
 	if (size_hints == NULL) {
 		*min_width = DBL_MIN;
@@ -448,11 +496,12 @@ static void handle_destroy(struct wl_listener *listener, void *data) {
 	wl_list_remove(&xwayland_view->set_title.link);
 	wl_list_remove(&xwayland_view->set_class.link);
 	wl_list_remove(&xwayland_view->set_role.link);
+	wl_list_remove(&xwayland_view->set_startup_id.link);
 	wl_list_remove(&xwayland_view->set_window_type.link);
 	wl_list_remove(&xwayland_view->set_hints.link);
 	wl_list_remove(&xwayland_view->set_decorations.link);
-	wl_list_remove(&xwayland_view->map.link);
-	wl_list_remove(&xwayland_view->unmap.link);
+	wl_list_remove(&xwayland_view->associate.link);
+	wl_list_remove(&xwayland_view->dissociate.link);
 	wl_list_remove(&xwayland_view->override_redirect.link);
 	view_begin_destroy(&xwayland_view->view);
 }
@@ -474,8 +523,8 @@ static void handle_unmap(struct wl_listener *listener, void *data) {
 static void handle_map(struct wl_listener *listener, void *data) {
 	struct sway_xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, map);
-	struct wlr_xwayland_surface *xsurface = data;
 	struct sway_view *view = &xwayland_view->view;
+	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
 
 	view->natural_width = xsurface->width;
 	view->natural_height = xsurface->height;
@@ -491,20 +540,30 @@ static void handle_map(struct wl_listener *listener, void *data) {
 	transaction_commit_dirty();
 }
 
+static void handle_dissociate(struct wl_listener *listener, void *data);
+
 static void handle_override_redirect(struct wl_listener *listener, void *data) {
 	struct sway_xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, override_redirect);
-	struct wlr_xwayland_surface *xsurface = data;
 	struct sway_view *view = &xwayland_view->view;
+	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
 
-	bool mapped = xsurface->mapped;
+	bool associated = xsurface->surface != NULL;
+	bool mapped = associated && xsurface->surface->mapped;
 	if (mapped) {
 		handle_unmap(&xwayland_view->unmap, NULL);
 	}
+	if (associated) {
+		handle_dissociate(&xwayland_view->dissociate, NULL);
+	}
 
 	handle_destroy(&xwayland_view->destroy, view);
 	xsurface->data = NULL;
+
 	struct sway_xwayland_unmanaged *unmanaged = create_unmanaged(xsurface);
+	if (associated) {
+		unmanaged_handle_associate(&unmanaged->associate, NULL);
+	}
 	if (mapped) {
 		unmanaged_handle_map(&unmanaged->map, xsurface);
 	}
@@ -516,7 +575,7 @@ static void handle_request_configure(struct wl_listener *listener, void *data) {
 	struct wlr_xwayland_surface_configure_event *ev = data;
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		wlr_xwayland_surface_configure(xsurface, ev->x, ev->y,
 			ev->width, ev->height);
 		return;
@@ -545,7 +604,7 @@ static void handle_request_fullscreen(struct wl_listener *listener, void *data)
 		wl_container_of(listener, xwayland_view, request_fullscreen);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 	container_set_fullscreen(view->container, xsurface->fullscreen);
@@ -559,7 +618,7 @@ static void handle_request_minimize(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, request_minimize);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 
@@ -574,10 +633,11 @@ static void handle_request_move(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, request_move);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
-	if (!container_is_floating(view->container)) {
+	if (!container_is_floating(view->container) ||
+			view->container->pending.fullscreen_mode) {
 		return;
 	}
 	struct sway_seat *seat = input_manager_current_seat();
@@ -589,7 +649,7 @@ static void handle_request_resize(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, request_resize);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 	if (!container_is_floating(view->container)) {
@@ -605,10 +665,10 @@ static void handle_request_activate(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, request_activate);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
-	view_request_activate(view);
+	view_request_activate(view, NULL);
 
 	transaction_commit_dirty();
 }
@@ -618,7 +678,7 @@ static void handle_set_title(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, set_title);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 	view_update_title(view, false);
@@ -630,7 +690,7 @@ static void handle_set_class(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, set_class);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 	view_execute_criteria(view);
@@ -641,18 +701,43 @@ static void handle_set_role(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, set_role);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 	view_execute_criteria(view);
 }
 
+static void handle_set_startup_id(struct wl_listener *listener, void *data) {
+	struct sway_xwayland_view *xwayland_view =
+		wl_container_of(listener, xwayland_view, set_startup_id);
+	struct sway_view *view = &xwayland_view->view;
+	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
+	if (xsurface->startup_id == NULL) {
+		return;
+	}
+
+	struct wlr_xdg_activation_token_v1 *token =
+		wlr_xdg_activation_v1_find_token(
+				server.xdg_activation_v1, xsurface->startup_id);
+	if (token == NULL) {
+		// Tried to activate with an unknown or expired token
+		return;
+	}
+
+	struct launcher_ctx *ctx = token->data;
+	if (token->data == NULL) {
+		// TODO: support external launchers in X
+		return;
+	}
+	view_assign_ctx(view, ctx);
+}
+
 static void handle_set_window_type(struct wl_listener *listener, void *data) {
 	struct sway_xwayland_view *xwayland_view =
 		wl_container_of(listener, xwayland_view, set_window_type);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
 	view_execute_criteria(view);
@@ -663,20 +748,39 @@ static void handle_set_hints(struct wl_listener *listener, void *data) {
 		wl_container_of(listener, xwayland_view, set_hints);
 	struct sway_view *view = &xwayland_view->view;
 	struct wlr_xwayland_surface *xsurface = view->wlr_xwayland_surface;
-	if (!xsurface->mapped) {
+	if (xsurface->surface == NULL || !xsurface->surface->mapped) {
 		return;
 	}
-	if (!xsurface->hints_urgency && view->urgent_timer) {
+	const bool hints_urgency = xcb_icccm_wm_hints_get_urgency(xsurface->hints);
+	if (!hints_urgency && view->urgent_timer) {
 		// The view is in the timeout period. We'll ignore the request to
 		// unset urgency so that the view remains urgent until the timer clears
 		// it.
 		return;
 	}
 	if (view->allow_request_urgent) {
-		view_set_urgent(view, (bool)xsurface->hints_urgency);
+		view_set_urgent(view, hints_urgency);
 	}
 }
 
+static void handle_associate(struct wl_listener *listener, void *data) {
+	struct sway_xwayland_view *xwayland_view =
+		wl_container_of(listener, xwayland_view, associate);
+	struct wlr_xwayland_surface *xsurface =
+		xwayland_view->view.wlr_xwayland_surface;
+	wl_signal_add(&xsurface->surface->events.unmap, &xwayland_view->unmap);
+	xwayland_view->unmap.notify = handle_unmap;
+	wl_signal_add(&xsurface->surface->events.map, &xwayland_view->map);
+	xwayland_view->map.notify = handle_map;
+}
+
+static void handle_dissociate(struct wl_listener *listener, void *data) {
+	struct sway_xwayland_view *xwayland_view =
+		wl_container_of(listener, xwayland_view, dissociate);
+	wl_list_remove(&xwayland_view->map.link);
+	wl_list_remove(&xwayland_view->unmap.link);
+}
+
 struct sway_view *view_from_wlr_xwayland_surface(
 		struct wlr_xwayland_surface *xsurface) {
 	return xsurface->data;
@@ -731,6 +835,10 @@ struct sway_xwayland_view *create_xwayland_view(struct wlr_xwayland_surface *xsu
 	wl_signal_add(&xsurface->events.set_role, &xwayland_view->set_role);
 	xwayland_view->set_role.notify = handle_set_role;
 
+	wl_signal_add(&xsurface->events.set_startup_id,
+			&xwayland_view->set_startup_id);
+	xwayland_view->set_startup_id.notify = handle_set_startup_id;
+
 	wl_signal_add(&xsurface->events.set_window_type,
 			&xwayland_view->set_window_type);
 	xwayland_view->set_window_type.notify = handle_set_window_type;
@@ -742,11 +850,11 @@ struct sway_xwayland_view *create_xwayland_view(struct wlr_xwayland_surface *xsu
 			&xwayland_view->set_decorations);
 	xwayland_view->set_decorations.notify = handle_set_decorations;
 
-	wl_signal_add(&xsurface->events.unmap, &xwayland_view->unmap);
-	xwayland_view->unmap.notify = handle_unmap;
+	wl_signal_add(&xsurface->events.associate, &xwayland_view->associate);
+	xwayland_view->associate.notify = handle_associate;
 
-	wl_signal_add(&xsurface->events.map, &xwayland_view->map);
-	xwayland_view->map.notify = handle_map;
+	wl_signal_add(&xsurface->events.dissociate, &xwayland_view->dissociate);
+	xwayland_view->dissociate.notify = handle_dissociate;
 
 	wl_signal_add(&xsurface->events.set_override_redirect,
 			&xwayland_view->override_redirect);
