@@ -7,7 +7,7 @@
 #include <time.h>
 #include <strings.h>
 #include <wlr/types/wlr_cursor.h>
-#include <wlr/types/wlr_idle.h>
+#include <wlr/types/wlr_cursor_shape_v1.h>
 #include <wlr/types/wlr_pointer.h>
 #include <wlr/types/wlr_touch.h>
 #include <wlr/types/wlr_tablet_v2.h>
@@ -53,12 +53,10 @@ static struct wlr_surface *layer_surface_at(struct sway_output *output,
 }
 
 static bool surface_is_xdg_popup(struct wlr_surface *surface) {
-    if (wlr_surface_is_xdg_surface(surface)) {
-        struct wlr_xdg_surface *xdg_surface =
-            wlr_xdg_surface_from_wlr_surface(surface);
-        return xdg_surface->role == WLR_XDG_SURFACE_ROLE_POPUP;
-    }
-    return false;
+	struct wlr_xdg_surface *xdg_surface =
+		wlr_xdg_surface_try_from_wlr_surface(surface);
+	return xdg_surface != NULL && xdg_surface->role == WLR_XDG_SURFACE_ROLE_POPUP &&
+		xdg_surface->popup != NULL;
 }
 
 static struct wlr_surface *layer_surface_popup_at(struct sway_output *output,
@@ -97,6 +95,24 @@ struct sway_node *node_at_coords(
 	double ox = lx, oy = ly;
 	wlr_output_layout_output_coords(root->output_layout, wlr_output, &ox, &oy);
 
+	if (server.session_lock.locked) {
+		if (server.session_lock.lock == NULL) {
+			return NULL;
+		}
+		struct wlr_session_lock_surface_v1 *lock_surf;
+		wl_list_for_each(lock_surf, &server.session_lock.lock->surfaces, link) {
+			if (lock_surf->output != wlr_output) {
+				continue;
+			}
+
+			*surface = wlr_surface_surface_at(lock_surf->surface, ox, oy, sx, sy);
+			if (*surface != NULL) {
+				return NULL;
+			}
+		}
+		return NULL;
+	}
+
 	// layer surfaces on the overlay layer are rendered on top
 	if ((*surface = layer_surface_at(output,
 				&output->layers[ZWLR_LAYER_SHELL_V1_LAYER_OVERLAY],
@@ -221,7 +237,7 @@ void cursor_update_image(struct sway_cursor *cursor,
 		// Try a node's resize edge
 		enum wlr_edges edge = find_resize_edge(node->sway_container, NULL, cursor);
 		if (edge == WLR_EDGE_NONE) {
-			cursor_set_image(cursor, "left_ptr", NULL);
+			cursor_set_image(cursor, "default", NULL);
 		} else if (container_is_floating(node->sway_container)) {
 			cursor_set_image(cursor, wlr_xcursor_get_resize_name(edge), NULL);
 		} else {
@@ -232,12 +248,12 @@ void cursor_update_image(struct sway_cursor *cursor,
 			}
 		}
 	} else {
-		cursor_set_image(cursor, "left_ptr", NULL);
+		cursor_set_image(cursor, "default", NULL);
 	}
 }
 
 static void cursor_hide(struct sway_cursor *cursor) {
-	wlr_cursor_set_image(cursor->cursor, NULL, 0, 0, 0, 0, 0, 0);
+	wlr_cursor_unset_image(cursor->cursor);
 	cursor->hidden = true;
 	wlr_seat_pointer_notify_clear_focus(cursor->seat->wlr_seat);
 }
@@ -349,7 +365,7 @@ void cursor_unhide(struct sway_cursor *cursor) {
 	wl_event_source_timer_update(cursor->hide_source, cursor_get_timeout(cursor));
 }
 
-static void pointer_motion(struct sway_cursor *cursor, uint32_t time_msec,
+void pointer_motion(struct sway_cursor *cursor, uint32_t time_msec,
 		struct wlr_input_device *device, double dx, double dy,
 		double dx_unaccel, double dy_unaccel) {
 	wlr_relative_pointer_manager_v1_send_relative_motion(
@@ -386,28 +402,29 @@ static void pointer_motion(struct sway_cursor *cursor, uint32_t time_msec,
 static void handle_pointer_motion_relative(
 		struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, motion);
-	struct wlr_event_pointer_motion *e = data;
-	cursor_handle_activity_from_device(cursor, e->device);
+	struct wlr_pointer_motion_event *e = data;
+	cursor_handle_activity_from_device(cursor, &e->pointer->base);
 
-	pointer_motion(cursor, e->time_msec, e->device, e->delta_x, e->delta_y,
-			e->unaccel_dx, e->unaccel_dy);
+	pointer_motion(cursor, e->time_msec, &e->pointer->base, e->delta_x,
+		e->delta_y, e->unaccel_dx, e->unaccel_dy);
 }
 
 static void handle_pointer_motion_absolute(
 		struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor =
 		wl_container_of(listener, cursor, motion_absolute);
-	struct wlr_event_pointer_motion_absolute *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_pointer_motion_absolute_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
 
 	double lx, ly;
-	wlr_cursor_absolute_to_layout_coords(cursor->cursor, event->device,
+	wlr_cursor_absolute_to_layout_coords(cursor->cursor, &event->pointer->base,
 			event->x, event->y, &lx, &ly);
 
 	double dx = lx - cursor->cursor->x;
 	double dy = ly - cursor->cursor->y;
 
-	pointer_motion(cursor, event->time_msec, event->device, dx, dy, dx, dy);
+	pointer_motion(cursor, event->time_msec, &event->pointer->base, dx, dy,
+		dx, dy);
 }
 
 void dispatch_cursor_button(struct sway_cursor *cursor,
@@ -422,7 +439,7 @@ void dispatch_cursor_button(struct sway_cursor *cursor,
 
 static void handle_pointer_button(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, button);
-	struct wlr_event_pointer_button *event = data;
+	struct wlr_pointer_button_event *event = data;
 
 	if (event->state == WLR_BUTTON_PRESSED) {
 		cursor->pressed_button_count++;
@@ -434,20 +451,20 @@ static void handle_pointer_button(struct wl_listener *listener, void *data) {
 		}
 	}
 
-	cursor_handle_activity_from_device(cursor, event->device);
-	dispatch_cursor_button(cursor, event->device,
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	dispatch_cursor_button(cursor, &event->pointer->base,
 			event->time_msec, event->button, event->state);
 }
 
 void dispatch_cursor_axis(struct sway_cursor *cursor,
-		struct wlr_event_pointer_axis *event) {
+		struct wlr_pointer_axis_event *event) {
 	seatop_pointer_axis(cursor->seat, event);
 }
 
 static void handle_pointer_axis(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, axis);
-	struct wlr_event_pointer_axis *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_pointer_axis_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
 	dispatch_cursor_axis(cursor, event);
 }
 
@@ -458,87 +475,74 @@ static void handle_pointer_frame(struct wl_listener *listener, void *data) {
 
 static void handle_touch_down(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, touch_down);
-	struct wlr_event_touch_down *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_touch_down_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->touch->base);
 	cursor_hide(cursor);
 
 	struct sway_seat *seat = cursor->seat;
-	struct wlr_seat *wlr_seat = seat->wlr_seat;
-	struct wlr_surface *surface = NULL;
 
 	double lx, ly;
-	wlr_cursor_absolute_to_layout_coords(cursor->cursor, event->device,
+	wlr_cursor_absolute_to_layout_coords(cursor->cursor, &event->touch->base,
 			event->x, event->y, &lx, &ly);
-	double sx, sy;
-	struct sway_node *focused_node = node_at_coords(seat, lx, ly, &surface, &sx, &sy);
 
 	seat->touch_id = event->touch_id;
 	seat->touch_x = lx;
 	seat->touch_y = ly;
 
-	if (surface && wlr_surface_accepts_touch(wlr_seat, surface)) {
-		if (seat_is_input_allowed(seat, surface)) {
-			wlr_seat_touch_notify_down(wlr_seat, surface, event->time_msec,
-					event->touch_id, sx, sy);
-
-			if (focused_node) {
-			    seat_set_focus(seat, focused_node);
-			}
-		}
-	} else if (!cursor->simulating_pointer_from_touch &&
-			(!surface || seat_is_input_allowed(seat, surface))) {
-		// Fallback to cursor simulation.
-		// The pointer_touch_id state is needed, so drags are not aborted when over
-		// a surface supporting touch and multi touch events don't interfere.
-		cursor->simulating_pointer_from_touch = true;
-		cursor->pointer_touch_id = seat->touch_id;
-		double dx, dy;
-		dx = lx - cursor->cursor->x;
-		dy = ly - cursor->cursor->y;
-		pointer_motion(cursor, event->time_msec, event->device, dx, dy, dx, dy);
-		dispatch_cursor_button(cursor, event->device, event->time_msec,
-				BTN_LEFT, WLR_BUTTON_PRESSED);
-	}
+	seatop_touch_down(seat, event, lx, ly);
 }
 
 static void handle_touch_up(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, touch_up);
-	struct wlr_event_touch_up *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_touch_up_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->touch->base);
 
-	struct wlr_seat *wlr_seat = cursor->seat->wlr_seat;
+	struct sway_seat *seat = cursor->seat;
 
 	if (cursor->simulating_pointer_from_touch) {
 		if (cursor->pointer_touch_id == cursor->seat->touch_id) {
 			cursor->pointer_touch_up = true;
-			dispatch_cursor_button(cursor, event->device, event->time_msec,
-					BTN_LEFT, WLR_BUTTON_RELEASED);
+			dispatch_cursor_button(cursor, &event->touch->base,
+				event->time_msec, BTN_LEFT, WLR_BUTTON_RELEASED);
 		}
 	} else {
-		wlr_seat_touch_notify_up(wlr_seat, event->time_msec, event->touch_id);
+		seatop_touch_up(seat, event);
+	}
+}
+
+static void handle_touch_cancel(struct wl_listener *listener, void *data) {
+	struct sway_cursor *cursor = wl_container_of(listener, cursor, touch_cancel);
+	struct wlr_touch_cancel_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->touch->base);
+
+	struct sway_seat *seat = cursor->seat;
+
+	if (cursor->simulating_pointer_from_touch) {
+		if (cursor->pointer_touch_id == cursor->seat->touch_id) {
+			cursor->pointer_touch_up = true;
+			dispatch_cursor_button(cursor, &event->touch->base,
+				event->time_msec, BTN_LEFT, WLR_BUTTON_RELEASED);
+		}
+	} else {
+		seatop_touch_cancel(seat, event);
 	}
 }
 
 static void handle_touch_motion(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor =
 		wl_container_of(listener, cursor, touch_motion);
-	struct wlr_event_touch_motion *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_touch_motion_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->touch->base);
 
 	struct sway_seat *seat = cursor->seat;
-	struct wlr_seat *wlr_seat = seat->wlr_seat;
-	struct wlr_surface *surface = NULL;
 
 	double lx, ly;
-	wlr_cursor_absolute_to_layout_coords(cursor->cursor, event->device,
+	wlr_cursor_absolute_to_layout_coords(cursor->cursor, &event->touch->base,
 			event->x, event->y, &lx, &ly);
-	double sx, sy;
-	node_at_coords(cursor->seat, lx, ly, &surface, &sx, &sy);
 
 	if (seat->touch_id == event->touch_id) {
 		seat->touch_x = lx;
 		seat->touch_y = ly;
-
 		struct sway_drag_icon *drag_icon;
 		wl_list_for_each(drag_icon, &root->drag_icons, link) {
 			if (drag_icon->seat == seat) {
@@ -552,11 +556,11 @@ static void handle_touch_motion(struct wl_listener *listener, void *data) {
 			double dx, dy;
 			dx = lx - cursor->cursor->x;
 			dy = ly - cursor->cursor->y;
-			pointer_motion(cursor, event->time_msec, event->device, dx, dy, dx, dy);
+			pointer_motion(cursor, event->time_msec, &event->touch->base,
+				dx, dy, dx, dy);
 		}
-	} else if (surface) {
-		wlr_seat_touch_notify_motion(wlr_seat, event->time_msec,
-			event->touch_id, sx, sy);
+	} else {
+		seatop_touch_motion(seat, event, lx, ly);
 	}
 }
 
@@ -591,14 +595,15 @@ static void apply_mapping_from_region(struct wlr_input_device *device,
 	double x1 = region->x1, x2 = region->x2;
 	double y1 = region->y1, y2 = region->y2;
 
-	if (region->mm) {
-		if (device->width_mm == 0 || device->height_mm == 0) {
+	if (region->mm && device->type == WLR_INPUT_DEVICE_TABLET_TOOL) {
+		struct wlr_tablet *tablet = wlr_tablet_from_input_device(device);
+		if (tablet->width_mm == 0 || tablet->height_mm == 0) {
 			return;
 		}
-		x1 /= device->width_mm;
-		x2 /= device->width_mm;
-		y1 /= device->height_mm;
-		y2 /= device->height_mm;
+		x1 /= tablet->width_mm;
+		x2 /= tablet->width_mm;
+		y1 /= tablet->height_mm;
+		y2 /= tablet->height_mm;
 	}
 
 	*x = apply_mapping_from_coord(x1, x2, *x);
@@ -660,8 +665,8 @@ static void handle_tablet_tool_position(struct sway_cursor *cursor,
 
 static void handle_tool_axis(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, tool_axis);
-	struct wlr_event_tablet_tool_axis *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_tablet_tool_axis_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->tablet->base);
 
 	struct sway_tablet_tool *sway_tool = event->tool->data;
 	if (!sway_tool) {
@@ -716,8 +721,8 @@ static void handle_tool_axis(struct wl_listener *listener, void *data) {
 
 static void handle_tool_tip(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, tool_tip);
-	struct wlr_event_tablet_tool_tip *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_tablet_tool_tip_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->tablet->base);
 
 	struct sway_tablet_tool *sway_tool = event->tool->data;
 	struct wlr_tablet_v2_tablet *tablet_v2 = sway_tool->tablet->tablet_v2;
@@ -732,7 +737,7 @@ static void handle_tool_tip(struct wl_listener *listener, void *data) {
 	if (cursor->simulating_pointer_from_tool_tip &&
 			event->state == WLR_TABLET_TOOL_TIP_UP) {
 		cursor->simulating_pointer_from_tool_tip = false;
-		dispatch_cursor_button(cursor, event->device, event->time_msec,
+		dispatch_cursor_button(cursor, &event->tablet->base, event->time_msec,
 			BTN_LEFT, WLR_BUTTON_RELEASED);
 		wlr_seat_pointer_notify_frame(cursor->seat->wlr_seat);
 	} else if (!surface || !wlr_surface_accepts_tablet_v2(tablet_v2, surface)) {
@@ -744,8 +749,8 @@ static void handle_tool_tip(struct wl_listener *listener, void *data) {
 				WLR_TABLET_TOOL_TIP_UP);
 		} else {
 			cursor->simulating_pointer_from_tool_tip = true;
-			dispatch_cursor_button(cursor, event->device, event->time_msec,
-				BTN_LEFT, WLR_BUTTON_PRESSED);
+			dispatch_cursor_button(cursor, &event->tablet->base,
+				event->time_msec, BTN_LEFT, WLR_BUTTON_PRESSED);
 			wlr_seat_pointer_notify_frame(cursor->seat->wlr_seat);
 		}
 	} else {
@@ -767,12 +772,13 @@ static struct sway_tablet *get_tablet_for_device(struct sway_cursor *cursor,
 static void handle_tool_proximity(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor =
 		wl_container_of(listener, cursor, tool_proximity);
-	struct wlr_event_tablet_tool_proximity *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_tablet_tool_proximity_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->tablet->base);
 
 	struct wlr_tablet_tool *tool = event->tool;
 	if (!tool->data) {
-		struct sway_tablet *tablet = get_tablet_for_device(cursor, event->device);
+		struct sway_tablet *tablet = get_tablet_for_device(cursor,
+			&event->tablet->base);
 		if (!tablet) {
 			sway_log(SWAY_ERROR, "no tablet for tablet tool");
 			return;
@@ -797,8 +803,8 @@ static void handle_tool_proximity(struct wl_listener *listener, void *data) {
 
 static void handle_tool_button(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, tool_button);
-	struct wlr_event_tablet_tool_button *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
+	struct wlr_tablet_tool_button_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->tablet->base);
 
 	struct sway_tablet_tool *sway_tool = event->tool->data;
 	if (!sway_tool) {
@@ -813,31 +819,71 @@ static void handle_tool_button(struct wl_listener *listener, void *data) {
 	node_at_coords(cursor->seat, cursor->cursor->x, cursor->cursor->y,
 		&surface, &sx, &sy);
 
-	if (!surface || !wlr_surface_accepts_tablet_v2(tablet_v2, surface)) {
+	// TODO: floating resize should support graphics tablet events
+	struct wlr_keyboard *keyboard = wlr_seat_get_keyboard(cursor->seat->wlr_seat);
+	uint32_t modifiers = keyboard ? wlr_keyboard_get_modifiers(keyboard) : 0;
+	bool mod_pressed = modifiers & config->floating_mod;
+
+	bool surface_supports_tablet_events =
+		surface && wlr_surface_accepts_tablet_v2(tablet_v2, surface);
+
+	// Simulate pointer when:
+	// 1. The modifier key is pressed, OR
+	// 2. The surface under the cursor does not support tablet events.
+	bool should_simulate_pointer = mod_pressed || !surface_supports_tablet_events;
+
+	// Similar to tool tip, we need to selectively simulate mouse events, but we
+	// want to make sure that it is always consistent. Because all tool buttons
+	// currently map to BTN_RIGHT, we need to keep count of how many tool
+	// buttons are currently pressed down so we can send consistent events.
+	//
+	// The logic follows:
+	// - If we are already simulating the pointer, we should continue to do so
+	//   until at least no tool button is held down.
+	// - If we should simulate the pointer and no tool button is currently held
+	//   down, begin simulating the pointer.
+	// - If neither of the above are true, send the tablet events.
+	if ((cursor->tool_buttons > 0 && cursor->simulating_pointer_from_tool_button)
+		|| (cursor->tool_buttons == 0 && should_simulate_pointer)) {
+		cursor->simulating_pointer_from_tool_button = true;
+
 		// TODO: the user may want to configure which tool buttons are mapped to
 		// which simulated pointer buttons
 		switch (event->state) {
 		case WLR_BUTTON_PRESSED:
 			if (cursor->tool_buttons == 0) {
-				dispatch_cursor_button(cursor, event->device,
+				dispatch_cursor_button(cursor, &event->tablet->base,
 						event->time_msec, BTN_RIGHT, event->state);
 			}
-			cursor->tool_buttons++;
 			break;
 		case WLR_BUTTON_RELEASED:
-			if (cursor->tool_buttons == 1) {
-				dispatch_cursor_button(cursor, event->device,
+			if (cursor->tool_buttons <= 1) {
+				dispatch_cursor_button(cursor, &event->tablet->base,
 						event->time_msec, BTN_RIGHT, event->state);
 			}
-			cursor->tool_buttons--;
 			break;
 		}
 		wlr_seat_pointer_notify_frame(cursor->seat->wlr_seat);
-		return;
+	} else {
+		cursor->simulating_pointer_from_tool_button = false;
+
+		wlr_tablet_v2_tablet_tool_notify_button(sway_tool->tablet_v2_tool,
+			event->button, (enum zwp_tablet_pad_v2_button_state)event->state);
 	}
 
-	wlr_tablet_v2_tablet_tool_notify_button(sway_tool->tablet_v2_tool,
-		event->button, (enum zwp_tablet_pad_v2_button_state)event->state);
+	// Update tool button count.
+	switch (event->state) {
+	case WLR_BUTTON_PRESSED:
+		cursor->tool_buttons++;
+		break;
+	case WLR_BUTTON_RELEASED:
+		if (cursor->tool_buttons == 0) {
+			sway_log(SWAY_ERROR, "inconsistent tablet tool button events");
+		} else {
+			cursor->tool_buttons--;
+		}
+		break;
+	}
 }
 
 static void check_constraint_region(struct sway_cursor *cursor) {
@@ -923,65 +969,68 @@ static void handle_request_pointer_set_cursor(struct wl_listener *listener,
 			event->hotspot_y, focused_client);
 }
 
+static void handle_pointer_hold_begin(struct wl_listener *listener, void *data) {
+	struct sway_cursor *cursor = wl_container_of(
+			listener, cursor, hold_begin);
+	struct wlr_pointer_hold_begin_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_hold_begin(cursor->seat, event);
+}
+
+static void handle_pointer_hold_end(struct wl_listener *listener, void *data) {
+	struct sway_cursor *cursor = wl_container_of(
+			listener, cursor, hold_end);
+	struct wlr_pointer_hold_end_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_hold_end(cursor->seat, event);
+}
+
 static void handle_pointer_pinch_begin(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(
 			listener, cursor, pinch_begin);
-	struct wlr_event_pointer_pinch_begin *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
-	wlr_pointer_gestures_v1_send_pinch_begin(
-			cursor->pointer_gestures, cursor->seat->wlr_seat,
-			event->time_msec, event->fingers);
+	struct wlr_pointer_pinch_begin_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_pinch_begin(cursor->seat, event);
 }
 
 static void handle_pointer_pinch_update(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(
 			listener, cursor, pinch_update);
-	struct wlr_event_pointer_pinch_update *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
-	wlr_pointer_gestures_v1_send_pinch_update(
-			cursor->pointer_gestures, cursor->seat->wlr_seat,
-			event->time_msec, event->dx, event->dy,
-			event->scale, event->rotation);
+	struct wlr_pointer_pinch_update_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_pinch_update(cursor->seat, event);
 }
 
 static void handle_pointer_pinch_end(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(
 			listener, cursor, pinch_end);
-	struct wlr_event_pointer_pinch_end *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
-	wlr_pointer_gestures_v1_send_pinch_end(
-			cursor->pointer_gestures, cursor->seat->wlr_seat,
-			event->time_msec, event->cancelled);
+	struct wlr_pointer_pinch_end_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_pinch_end(cursor->seat, event);
 }
 
 static void handle_pointer_swipe_begin(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(
 			listener, cursor, swipe_begin);
-	struct wlr_event_pointer_swipe_begin *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
-	wlr_pointer_gestures_v1_send_swipe_begin(
-			cursor->pointer_gestures, cursor->seat->wlr_seat,
-			event->time_msec, event->fingers);
+	struct wlr_pointer_swipe_begin_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_swipe_begin(cursor->seat, event);
 }
 
 static void handle_pointer_swipe_update(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(
 			listener, cursor, swipe_update);
-	struct wlr_event_pointer_swipe_update *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
-	wlr_pointer_gestures_v1_send_swipe_update(
-			cursor->pointer_gestures, cursor->seat->wlr_seat,
-			event->time_msec, event->dx, event->dy);
+	struct wlr_pointer_swipe_update_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_swipe_update(cursor->seat, event);
 }
 
 static void handle_pointer_swipe_end(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(
 			listener, cursor, swipe_end);
-	struct wlr_event_pointer_swipe_end *event = data;
-	cursor_handle_activity_from_device(cursor, event->device);
-	wlr_pointer_gestures_v1_send_swipe_end(
-			cursor->pointer_gestures, cursor->seat->wlr_seat,
-			event->time_msec, event->cancelled);
+	struct wlr_pointer_swipe_end_event *event = data;
+	cursor_handle_activity_from_device(cursor, &event->pointer->base);
+	seatop_swipe_end(cursor->seat, event);
 }
 
 static void handle_image_surface_destroy(struct wl_listener *listener,
@@ -1020,10 +1069,9 @@ void cursor_set_image(struct sway_cursor *cursor, const char *image,
 	}
 
 	if (!image) {
-		wlr_cursor_set_image(cursor->cursor, NULL, 0, 0, 0, 0, 0, 0);
+		wlr_cursor_unset_image(cursor->cursor);
 	} else if (!current_image || strcmp(current_image, image) != 0) {
-		wlr_xcursor_manager_set_cursor_image(cursor->xcursor_manager, image,
-				cursor->cursor);
+		wlr_cursor_set_xcursor(cursor->cursor, cursor->xcursor_manager, image);
 	}
 }
 
@@ -1055,6 +1103,8 @@ void sway_cursor_destroy(struct sway_cursor *cursor) {
 	wl_event_source_remove(cursor->hide_source);
 
 	wl_list_remove(&cursor->image_surface_destroy.link);
+	wl_list_remove(&cursor->hold_begin.link);
+	wl_list_remove(&cursor->hold_end.link);
 	wl_list_remove(&cursor->pinch_begin.link);
 	wl_list_remove(&cursor->pinch_update.link);
 	wl_list_remove(&cursor->pinch_end.link);
@@ -1068,6 +1118,7 @@ void sway_cursor_destroy(struct sway_cursor *cursor) {
 	wl_list_remove(&cursor->frame.link);
 	wl_list_remove(&cursor->touch_down.link);
 	wl_list_remove(&cursor->touch_up.link);
+	wl_list_remove(&cursor->touch_cancel.link);
 	wl_list_remove(&cursor->touch_motion.link);
 	wl_list_remove(&cursor->touch_frame.link);
 	wl_list_remove(&cursor->tool_axis.link);
@@ -1104,19 +1155,24 @@ struct sway_cursor *sway_cursor_create(struct sway_seat *seat) {
 	wl_list_init(&cursor->image_surface_destroy.link);
 	cursor->image_surface_destroy.notify = handle_image_surface_destroy;
 
-	cursor->pointer_gestures = wlr_pointer_gestures_v1_create(server.wl_display);
-	cursor->pinch_begin.notify = handle_pointer_pinch_begin;
+	wl_signal_add(&wlr_cursor->events.hold_begin, &cursor->hold_begin);
+	cursor->hold_begin.notify = handle_pointer_hold_begin;
+	wl_signal_add(&wlr_cursor->events.hold_end, &cursor->hold_end);
+	cursor->hold_end.notify = handle_pointer_hold_end;
+
 	wl_signal_add(&wlr_cursor->events.pinch_begin, &cursor->pinch_begin);
-	cursor->pinch_update.notify = handle_pointer_pinch_update;
+	cursor->pinch_begin.notify = handle_pointer_pinch_begin;
 	wl_signal_add(&wlr_cursor->events.pinch_update, &cursor->pinch_update);
-	cursor->pinch_end.notify = handle_pointer_pinch_end;
+	cursor->pinch_update.notify = handle_pointer_pinch_update;
 	wl_signal_add(&wlr_cursor->events.pinch_end, &cursor->pinch_end);
-	cursor->swipe_begin.notify = handle_pointer_swipe_begin;
+	cursor->pinch_end.notify = handle_pointer_pinch_end;
+
 	wl_signal_add(&wlr_cursor->events.swipe_begin, &cursor->swipe_begin);
-	cursor->swipe_update.notify = handle_pointer_swipe_update;
+	cursor->swipe_begin.notify = handle_pointer_swipe_begin;
 	wl_signal_add(&wlr_cursor->events.swipe_update, &cursor->swipe_update);
-	cursor->swipe_end.notify = handle_pointer_swipe_end;
+	cursor->swipe_update.notify = handle_pointer_swipe_update;
 	wl_signal_add(&wlr_cursor->events.swipe_end, &cursor->swipe_end);
+	cursor->swipe_end.notify = handle_pointer_swipe_end;
 
 	// input events
 	wl_signal_add(&wlr_cursor->events.motion, &cursor->motion);
@@ -1141,6 +1197,9 @@ struct sway_cursor *sway_cursor_create(struct sway_seat *seat) {
 	wl_signal_add(&wlr_cursor->events.touch_up, &cursor->touch_up);
 	cursor->touch_up.notify = handle_touch_up;
 
+	wl_signal_add(&wlr_cursor->events.touch_cancel, &cursor->touch_cancel);
+	cursor->touch_cancel.notify = handle_touch_cancel;
+
 	wl_signal_add(&wlr_cursor->events.touch_motion,
 		&cursor->touch_motion);
 	cursor->touch_motion.notify = handle_touch_motion;
@@ -1233,11 +1292,7 @@ uint32_t get_mouse_bindsym(const char *name, char **error) {
 		// Get event code from name
 		int code = libevdev_event_code_from_name(EV_KEY, name);
 		if (code == -1) {
-			size_t len = snprintf(NULL, 0, "Unknown event %s", name) + 1;
-			*error = malloc(len);
-			if (*error) {
-				snprintf(*error, len, "Unknown event %s", name);
-			}
+			*error = format_str("Unknown event %s", name);
 			return 0;
 		}
 		return code;
@@ -1259,13 +1314,8 @@ uint32_t get_mouse_bindcode(const char *name, char **error) {
 	}
 	const char *event = libevdev_event_code_get_name(EV_KEY, code);
 	if (!event || strncmp(event, "BTN_", strlen("BTN_")) != 0) {
-		size_t len = snprintf(NULL, 0, "Event code %d (%s) is not a button",
-				code, event ? event : "(null)") + 1;
-		*error = malloc(len);
-		if (*error) {
-			snprintf(*error, len, "Event code %d (%s) is not a button",
-					code, event ? event : "(null)");
-		}
+		*error = format_str("Event code %d (%s) is not a button",
+			code, event ? event : "(null)");
 		return 0;
 	}
 	return code;
@@ -1304,6 +1354,10 @@ static void warp_to_constraint_cursor_hint(struct sway_cursor *cursor) {
 		double sy = constraint->current.cursor_hint.y;
 
 		struct sway_view *view = view_from_wlr_surface(constraint->surface);
+		if (!view) {
+			return;
+		}
+
 		struct sway_container *con = view->container;
 
 		double lx = sx + con->pending.content_x - view->geometry.x;
@@ -1354,12 +1408,9 @@ void handle_pointer_constraint(struct wl_listener *listener, void *data) {
 	sway_constraint->destroy.notify = handle_constraint_destroy;
 	wl_signal_add(&constraint->events.destroy, &sway_constraint->destroy);
 
-	struct sway_node *focus = seat_get_focus(seat);
-	if (focus && node_is_view(focus)) {
-		struct wlr_surface *surface = focus->sway_container->view->surface;
-		if (surface == constraint->surface) {
-			sway_cursor_constrain(seat->cursor, constraint);
-		}
+	struct wlr_surface *surface = seat->wlr_seat->keyboard_state.focused_surface;
+	if (surface && surface == constraint->surface) {
+		sway_cursor_constrain(seat->cursor, constraint);
 	}
 }
 
@@ -1417,3 +1468,26 @@ void sway_cursor_constrain(struct sway_cursor *cursor,
 	wl_signal_add(&constraint->surface->events.commit,
 		&cursor->constraint_commit);
 }
+
+void handle_request_set_cursor_shape(struct wl_listener *listener, void *data) {
+	const struct wlr_cursor_shape_manager_v1_request_set_shape_event *event = data;
+	struct sway_seat *seat = event->seat_client->seat->data;
+
+	if (!seatop_allows_set_cursor(seat)) {
+		return;
+	}
+
+	struct wl_client *focused_client = NULL;
+	struct wlr_surface *focused_surface = seat->wlr_seat->pointer_state.focused_surface;
+	if (focused_surface != NULL) {
+		focused_client = wl_resource_get_client(focused_surface->resource);
+	}
+
+	// TODO: check cursor mode
+	if (focused_client == NULL || event->seat_client->client != focused_client) {
+		sway_log(SWAY_DEBUG, "denying request to set cursor from unfocused client");
+		return;
+	}
+
+	cursor_set_image(seat->cursor, wlr_cursor_shape_v1_name(event->shape), focused_client);
+}
