@@ -77,8 +77,6 @@ static void handle_im_grab_keyboard(struct wl_listener *listener, void *data) {
 	struct wlr_keyboard *active_keyboard = wlr_seat_get_keyboard(relay->seat->wlr_seat);
 	wlr_input_method_keyboard_grab_v2_set_keyboard(keyboard_grab,
 		active_keyboard);
-	wlr_input_method_keyboard_grab_v2_send_modifiers(keyboard_grab,
-		&active_keyboard->modifiers);
 
 	wl_signal_add(&keyboard_grab->events.destroy,
 		&relay->input_method_keyboard_grab_destroy);
