@@ -3,7 +3,8 @@
 #include <json.h>
 #include <libevdev/libevdev.h>
 #include <stdio.h>
-#include <wlr/backend/libinput.h>
+#include <wlr/config.h>
+#include <wlr/types/wlr_content_type_v1.h>
 #include <wlr/types/wlr_output.h>
 #include <xkbcommon/xkbcommon.h>
 #include "config.h"
@@ -20,6 +21,10 @@
 #include "wlr-layer-shell-unstable-v1-protocol.h"
 #include "sway/desktop/idle_inhibit_v1.h"
 
+#if WLR_HAS_LIBINPUT_BACKEND
+#include <wlr/backend/libinput.h>
+#endif
+
 static const int i3_output_id = INT32_MAX;
 static const int i3_scratch_id = INT32_MAX - 1;
 
@@ -112,12 +117,43 @@ static const char *ipc_json_output_adaptive_sync_status_description(
 		return "disabled";
 	case WLR_OUTPUT_ADAPTIVE_SYNC_ENABLED:
 		return "enabled";
-	case WLR_OUTPUT_ADAPTIVE_SYNC_UNKNOWN:
-		return "unknown";
 	}
 	return NULL;
 }
 
+static const char *ipc_json_output_mode_aspect_ratio_description(
+		enum wlr_output_mode_aspect_ratio aspect_ratio) {
+	switch (aspect_ratio) {
+	case WLR_OUTPUT_MODE_ASPECT_RATIO_NONE:
+		return "none";
+	case WLR_OUTPUT_MODE_ASPECT_RATIO_4_3:
+		return "4:3";
+	case WLR_OUTPUT_MODE_ASPECT_RATIO_16_9:
+		return "16:9";
+	case WLR_OUTPUT_MODE_ASPECT_RATIO_64_27:
+		return "64:27";
+	case WLR_OUTPUT_MODE_ASPECT_RATIO_256_135:
+		return "256:135";
+	}
+	return NULL;
+}
+
+static json_object *ipc_json_output_mode_description(
+		const struct wlr_output_mode *mode) {
+	const char *pic_ar =
+		ipc_json_output_mode_aspect_ratio_description(mode->picture_aspect_ratio);
+	json_object *mode_object = json_object_new_object();
+	json_object_object_add(mode_object, "width",
+		json_object_new_int(mode->width));
+	json_object_object_add(mode_object, "height",
+		json_object_new_int(mode->height));
+	json_object_object_add(mode_object, "refresh",
+		json_object_new_int(mode->refresh));
+	json_object_object_add(mode_object, "picture_aspect_ratio",
+		json_object_new_string(pic_ar));
+	return mode_object;
+}
+
 #if HAVE_XWAYLAND
 static const char *ipc_json_xwindow_type_description(struct sway_view *view) {
 	struct wlr_xwayland_surface *surface = view->wlr_xwayland_surface;
@@ -170,6 +206,20 @@ static const char *ipc_json_user_idle_inhibitor_description(enum sway_idle_inhib
 	return NULL;
 }
 
+static const char *ipc_json_content_type_description(enum wp_content_type_v1_type type) {
+	switch (type) {
+	case WP_CONTENT_TYPE_V1_TYPE_NONE:
+		return "none";
+	case WP_CONTENT_TYPE_V1_TYPE_PHOTO:
+		return "photo";
+	case WP_CONTENT_TYPE_V1_TYPE_VIDEO:
+		return "video";
+	case WP_CONTENT_TYPE_V1_TYPE_GAME:
+		return "game";
+	}
+	return NULL;
+}
+
 json_object *ipc_json_get_version(void) {
 	int major = 0, minor = 0, patch = 0;
 	json_object *version = json_object_new_object();
@@ -242,23 +292,50 @@ static json_object *ipc_json_create_node(int id, const char* type, char *name,
 	return object;
 }
 
+static void ipc_json_describe_wlr_output(struct wlr_output *wlr_output, json_object *object) {
+	json_object_object_add(object, "primary", json_object_new_boolean(false));
+	json_object_object_add(object, "make",
+			json_object_new_string(wlr_output->make ? wlr_output->make : "Unknown"));
+	json_object_object_add(object, "model",
+			json_object_new_string(wlr_output->model ? wlr_output->model : "Unknown"));
+	json_object_object_add(object, "serial",
+			json_object_new_string(wlr_output->serial ? wlr_output->serial : "Unknown"));
+
+	json_object *modes_array = json_object_new_array();
+	struct wlr_output_mode *mode;
+	wl_list_for_each(mode, &wlr_output->modes, link) {
+		json_object *mode_object = json_object_new_object();
+		json_object_object_add(mode_object, "width",
+			json_object_new_int(mode->width));
+		json_object_object_add(mode_object, "height",
+			json_object_new_int(mode->height));
+		json_object_object_add(mode_object, "refresh",
+			json_object_new_int(mode->refresh));
+		json_object_array_add(modes_array, mode_object);
+	}
+	json_object_object_add(object, "modes", modes_array);
+}
+
 static void ipc_json_describe_output(struct sway_output *output,
 		json_object *object) {
+	ipc_json_describe_wlr_output(output->wlr_output, object);
+}
+
+static void ipc_json_describe_enabled_output(struct sway_output *output,
+		json_object *object) {
+	ipc_json_describe_output(output, object);
+
 	struct wlr_output *wlr_output = output->wlr_output;
+	json_object_object_add(object, "non_desktop", json_object_new_boolean(false));
 	json_object_object_add(object, "active", json_object_new_boolean(true));
 	json_object_object_add(object, "dpms",
 			json_object_new_boolean(wlr_output->enabled));
-	json_object_object_add(object, "primary", json_object_new_boolean(false));
+	json_object_object_add(object, "power",
+			json_object_new_boolean(wlr_output->enabled));
 	json_object_object_add(object, "layout", json_object_new_string("output"));
 	json_object_object_add(object, "orientation",
 			json_object_new_string(
 				ipc_json_orientation_description(L_NONE)));
-	json_object_object_add(object, "make",
-			json_object_new_string(wlr_output->make));
-	json_object_object_add(object, "model",
-			json_object_new_string(wlr_output->model));
-	json_object_object_add(object, "serial",
-			json_object_new_string(wlr_output->serial));
 	json_object_object_add(object, "scale",
 			json_object_new_double(wlr_output->scale));
 	json_object_object_add(object, "scale_filter",
@@ -283,25 +360,26 @@ static void ipc_json_describe_output(struct sway_output *output,
 	json_object *modes_array = json_object_new_array();
 	struct wlr_output_mode *mode;
 	wl_list_for_each(mode, &wlr_output->modes, link) {
-		json_object *mode_object = json_object_new_object();
-		json_object_object_add(mode_object, "width",
-			json_object_new_int(mode->width));
-		json_object_object_add(mode_object, "height",
-			json_object_new_int(mode->height));
-		json_object_object_add(mode_object, "refresh",
-			json_object_new_int(mode->refresh));
+		json_object *mode_object =
+			ipc_json_output_mode_description(mode);
 		json_object_array_add(modes_array, mode_object);
 	}
 
 	json_object_object_add(object, "modes", modes_array);
 
-	json_object *current_mode_object = json_object_new_object();
-	json_object_object_add(current_mode_object, "width",
-		json_object_new_int(wlr_output->width));
-	json_object_object_add(current_mode_object, "height",
-		json_object_new_int(wlr_output->height));
-	json_object_object_add(current_mode_object, "refresh",
-		json_object_new_int(wlr_output->refresh));
+	json_object *current_mode_object;
+	if (wlr_output->current_mode != NULL) {
+		current_mode_object =
+			ipc_json_output_mode_description(wlr_output->current_mode);
+	} else {
+		current_mode_object = json_object_new_object();
+		json_object_object_add(current_mode_object, "width",
+			json_object_new_int(wlr_output->width));
+		json_object_object_add(current_mode_object, "height",
+			json_object_new_int(wlr_output->height));
+		json_object_object_add(current_mode_object, "refresh",
+			json_object_new_int(wlr_output->refresh));
+	}
 	json_object_object_add(object, "current_mode", current_mode_object);
 
 	struct sway_node *parent = node_get_parent(&output->node);
@@ -325,33 +403,15 @@ json_object *ipc_json_describe_disabled_output(struct sway_output *output) {
 
 	json_object *object = json_object_new_object();
 
+	ipc_json_describe_output(output, object);
+
+	json_object_object_add(object, "non_desktop", json_object_new_boolean(false));
 	json_object_object_add(object, "type", json_object_new_string("output"));
 	json_object_object_add(object, "name",
 			json_object_new_string(wlr_output->name));
 	json_object_object_add(object, "active", json_object_new_boolean(false));
 	json_object_object_add(object, "dpms", json_object_new_boolean(false));
-	json_object_object_add(object, "primary", json_object_new_boolean(false));
-	json_object_object_add(object, "make",
-			json_object_new_string(wlr_output->make));
-	json_object_object_add(object, "model",
-			json_object_new_string(wlr_output->model));
-	json_object_object_add(object, "serial",
-			json_object_new_string(wlr_output->serial));
-
-	json_object *modes_array = json_object_new_array();
-	struct wlr_output_mode *mode;
-	wl_list_for_each(mode, &wlr_output->modes, link) {
-		json_object *mode_object = json_object_new_object();
-		json_object_object_add(mode_object, "width",
-			json_object_new_int(mode->width));
-		json_object_object_add(mode_object, "height",
-			json_object_new_int(mode->height));
-		json_object_object_add(mode_object, "refresh",
-			json_object_new_int(mode->refresh));
-		json_object_array_add(modes_array, mode_object);
-	}
-
-	json_object_object_add(object, "modes", modes_array);
+	json_object_object_add(object, "power", json_object_new_boolean(false));
 
 	json_object_object_add(object, "current_workspace", NULL);
 
@@ -367,6 +427,21 @@ json_object *ipc_json_describe_disabled_output(struct sway_output *output) {
 	return object;
 }
 
+json_object *ipc_json_describe_non_desktop_output(struct sway_output_non_desktop *output) {
+	struct wlr_output *wlr_output = output->wlr_output;
+
+	json_object *object = json_object_new_object();
+
+	ipc_json_describe_wlr_output(wlr_output, object);
+
+	json_object_object_add(object, "non_desktop", json_object_new_boolean(true));
+	json_object_object_add(object, "type", json_object_new_string("output"));
+	json_object_object_add(object, "name",
+				json_object_new_string(wlr_output->name));
+
+	return object;
+}
+
 static json_object *ipc_json_describe_scratchpad_output(void) {
 	struct wlr_box box;
 	root_get_box(root, &box);
@@ -453,7 +528,9 @@ static void ipc_json_describe_workspace(struct sway_workspace *workspace,
 
 static void get_deco_rect(struct sway_container *c, struct wlr_box *deco_rect) {
 	enum sway_container_layout parent_layout = container_parent_layout(c);
-	bool tab_or_stack = parent_layout == L_TABBED || parent_layout == L_STACKED;
+	list_t *siblings = container_get_siblings(c);
+	bool tab_or_stack = (parent_layout == L_TABBED || parent_layout == L_STACKED)
+		&& ((siblings && siblings->length > 1) || !config->hide_lone_tab);
 	if (((!tab_or_stack || container_is_floating(c)) &&
 				c->current.border != B_NORMAL) ||
 			c->pending.fullscreen_mode != FULLSCREEN_NONE ||
@@ -500,7 +577,7 @@ static void ipc_json_describe_view(struct sway_container *c, json_object *object
 
 	struct wlr_box window_box = {
 		c->pending.content_x - c->pending.x,
-		(c->current.border == B_PIXEL) ? c->current.border_thickness : 0,
+		(c->current.border == B_PIXEL) ? c->pending.content_y - c->pending.y : 0,
 		c->pending.content_width,
 		c->pending.content_height
 	};
@@ -544,6 +621,16 @@ static void ipc_json_describe_view(struct sway_container *c, json_object *object
 
 	json_object_object_add(object, "idle_inhibitors", idle_inhibitors);
 
+	enum wp_content_type_v1_type content_type = WP_CONTENT_TYPE_V1_TYPE_NONE;
+	if (c->view->surface != NULL) {
+		content_type = wlr_surface_get_content_type_v1(server.content_type_manager_v1,
+			c->view->surface);
+	}
+	if (content_type != WP_CONTENT_TYPE_V1_TYPE_NONE) {
+		json_object_object_add(object, "content_type",
+			json_object_new_string(ipc_json_content_type_description(content_type)));
+	}
+
 #if HAVE_XWAYLAND
 	if (c->view->type == SWAY_VIEW_XWAYLAND) {
 		json_object_object_add(object, "window",
@@ -706,7 +793,7 @@ json_object *ipc_json_describe_node(struct sway_node *node) {
 	case N_ROOT:
 		break;
 	case N_OUTPUT:
-		ipc_json_describe_output(node->sway_output, object);
+		ipc_json_describe_enabled_output(node->sway_output, object);
 		break;
 	case N_CONTAINER:
 		ipc_json_describe_container(node->sway_container, object);
@@ -764,6 +851,7 @@ json_object *ipc_json_describe_node_recursive(struct sway_node *node) {
 	return object;
 }
 
+#if WLR_HAS_LIBINPUT_BACKEND
 static json_object *describe_libinput_device(struct libinput_device *device) {
 	json_object *object = json_object_new_object();
 
@@ -847,6 +935,11 @@ static json_object *describe_libinput_device(struct libinput_device *device) {
 		case LIBINPUT_CONFIG_ACCEL_PROFILE_ADAPTIVE:
 			accel_profile = "adaptive";
 			break;
+#if HAVE_LIBINPUT_CONFIG_ACCEL_PROFILE_CUSTOM
+		case LIBINPUT_CONFIG_ACCEL_PROFILE_CUSTOM:
+			accel_profile = "custom";
+			break;
+#endif
 		}
 		json_object_object_add(object, "accel_profile",
 				json_object_new_string(accel_profile));
@@ -926,6 +1019,17 @@ static json_object *describe_libinput_device(struct libinput_device *device) {
 			uint32_t button = libinput_device_config_scroll_get_button(device);
 			json_object_object_add(object, "scroll_button",
 					json_object_new_int(button));
+			const char *lock = "unknown";
+			switch (libinput_device_config_scroll_get_button_lock(device)) {
+			case LIBINPUT_CONFIG_SCROLL_BUTTON_LOCK_ENABLED:
+				lock = "enabled";
+				break;
+			case LIBINPUT_CONFIG_SCROLL_BUTTON_LOCK_DISABLED:
+				lock = "disabled";
+				break;
+			}
+			json_object_object_add(object, "scroll_button_lock",
+					json_object_new_string(lock));
 		}
 	}
 
@@ -942,6 +1046,19 @@ static json_object *describe_libinput_device(struct libinput_device *device) {
 		json_object_object_add(object, "dwt", json_object_new_string(dwt));
 	}
 
+	if (libinput_device_config_dwtp_is_available(device)) {
+		const char *dwtp = "unknown";
+		switch (libinput_device_config_dwtp_get_enabled(device)) {
+		case LIBINPUT_CONFIG_DWTP_ENABLED:
+			dwtp = "enabled";
+			break;
+		case LIBINPUT_CONFIG_DWTP_DISABLED:
+			dwtp = "disabled";
+			break;
+		}
+		json_object_object_add(object, "dwtp", json_object_new_string(dwtp));
+	}
+
 	if (libinput_device_config_calibration_has_matrix(device)) {
 		float matrix[6];
 		libinput_device_config_calibration_get_matrix(device, matrix);
@@ -956,6 +1073,7 @@ static json_object *describe_libinput_device(struct libinput_device *device) {
 
 	return object;
 }
+#endif
 
 json_object *ipc_json_describe_input(struct sway_input_device *device) {
 	if (!(sway_assert(device, "Device must not be null"))) {
@@ -977,10 +1095,16 @@ json_object *ipc_json_describe_input(struct sway_input_device *device) {
 			input_device_get_type(device)));
 
 	if (device->wlr_device->type == WLR_INPUT_DEVICE_KEYBOARD) {
-		struct wlr_keyboard *keyboard = device->wlr_device->keyboard;
+		struct wlr_keyboard *keyboard =
+			wlr_keyboard_from_input_device(device->wlr_device);
 		struct xkb_keymap *keymap = keyboard->keymap;
 		struct xkb_state *state = keyboard->xkb_state;
 
+		json_object_object_add(object, "repeat_delay", 
+			json_object_new_int(keyboard->repeat_info.delay));
+		json_object_object_add(object, "repeat_rate", 
+			json_object_new_int(keyboard->repeat_info.rate));
+
 		json_object *layouts_arr = json_object_new_array();
 		json_object_object_add(object, "xkb_layout_names", layouts_arr);
 
@@ -1005,20 +1129,22 @@ json_object *ipc_json_describe_input(struct sway_input_device *device) {
 	if (device->wlr_device->type == WLR_INPUT_DEVICE_POINTER) {
 		struct input_config *ic = input_device_get_config(device);
 		float scroll_factor = 1.0f;
-		if (ic != NULL && !isnan(ic->scroll_factor) && 
+		if (ic != NULL && !isnan(ic->scroll_factor) &&
 				ic->scroll_factor != FLT_MIN) {
 			scroll_factor = ic->scroll_factor;
 		}
-		json_object_object_add(object, "scroll_factor", 
+		json_object_object_add(object, "scroll_factor",
 				json_object_new_double(scroll_factor));
 	}
 
+#if WLR_HAS_LIBINPUT_BACKEND
 	if (wlr_input_device_is_libinput(device->wlr_device)) {
 		struct libinput_device *libinput_dev;
 		libinput_dev = wlr_libinput_get_device_handle(device->wlr_device);
 		json_object_object_add(object, "libinput",
 				describe_libinput_device(libinput_dev));
 	}
+#endif
 
 	return object;
 }
