@@ -56,7 +56,6 @@ struct ipc_client {
 	enum ipc_command_type pending_type;
 };
 
-struct sockaddr_un *ipc_user_sockaddr(void);
 int ipc_handle_connection(int fd, uint32_t mask, void *data);
 int ipc_client_handle_readable(int client_fd, uint32_t mask, void *data);
 int ipc_client_handle_writable(int client_fd, uint32_t mask, void *data);
@@ -150,7 +149,6 @@ struct sockaddr_un *ipc_user_sockaddr(void) {
 int ipc_handle_connection(int fd, uint32_t mask, void *data) {
 	(void) fd;
 	struct sway_server *server = data;
-	sway_log(SWAY_DEBUG, "Event on IPC listening socket");
 	assert(mask == WL_EVENT_READABLE);
 
 	int client_fd = accept(ipc_socket, NULL, NULL);
@@ -211,13 +209,10 @@ int ipc_client_handle_readable(int client_fd, uint32_t mask, void *data) {
 	}
 
 	if (mask & WL_EVENT_HANGUP) {
-		sway_log(SWAY_DEBUG, "Client %d hung up", client->fd);
 		ipc_client_disconnect(client);
 		return 0;
 	}
 
-	sway_log(SWAY_DEBUG, "Client %d readable", client->fd);
-
 	int read_available;
 	if (ioctl(client_fd, FIONREAD, &read_available) == -1) {
 		sway_log_errno(SWAY_INFO, "Unable to read IPC socket buffer size");
@@ -513,6 +508,20 @@ void ipc_event_input(const char *change, struct sway_input_device *device) {
 	json_object_put(json);
 }
 
+void ipc_event_output(void) {
+	if (!ipc_has_event_listeners(IPC_EVENT_OUTPUT)) {
+		return;
+	}
+	sway_log(SWAY_DEBUG, "Sending output event");
+
+	json_object *json = json_object_new_object();
+	json_object_object_add(json, "change", json_object_new_string("unspecified"));
+
+	const char *json_string = json_object_to_json_string(json);
+	ipc_send_event(json_string, IPC_EVENT_OUTPUT);
+	json_object_put(json);
+}
+
 int ipc_client_handle_writable(int client_fd, uint32_t mask, void *data) {
 	struct ipc_client *client = data;
 
@@ -523,7 +532,6 @@ int ipc_client_handle_writable(int client_fd, uint32_t mask, void *data) {
 	}
 
 	if (mask & WL_EVENT_HANGUP) {
-		sway_log(SWAY_DEBUG, "Client %d hung up", client->fd);
 		ipc_client_disconnect(client);
 		return 0;
 	}
@@ -532,8 +540,6 @@ int ipc_client_handle_writable(int client_fd, uint32_t mask, void *data) {
 		return 0;
 	}
 
-	sway_log(SWAY_DEBUG, "Client %d writable", client->fd);
-
 	ssize_t written = write(client->fd, client->write_buffer, client->write_buffer_len);
 
 	if (written == -1 && errno == EAGAIN) {
@@ -692,6 +698,12 @@ void ipc_client_handle_command(struct ipc_client *client, uint32_t payload_lengt
 						ipc_json_describe_disabled_output(output));
 			}
 		}
+
+		for (int i = 0; i < root->non_desktop_outputs->length; i++) {
+			struct sway_output_non_desktop *non_desktop_output = root->non_desktop_outputs->items[i];
+			json_object_array_add(outputs, ipc_json_describe_non_desktop_output(non_desktop_output));
+		}
+
 		const char *json_string = json_object_to_json_string(outputs);
 		ipc_send_reply(client, payload_type, json_string,
 			(uint32_t)strlen(json_string));
@@ -727,6 +739,8 @@ void ipc_client_handle_command(struct ipc_client *client, uint32_t payload_lengt
 			const char *event_type = json_object_get_string(json_object_array_get_idx(request, i));
 			if (strcmp(event_type, "workspace") == 0) {
 				client->subscribed_events |= event_mask(IPC_EVENT_WORKSPACE);
+			} else if (strcmp(event_type, "output") == 0) {
+				client->subscribed_events |= event_mask(IPC_EVENT_OUTPUT);
 			} else if (strcmp(event_type, "barconfig_update") == 0) {
 				client->subscribed_events |= event_mask(IPC_EVENT_BARCONFIG_UPDATE);
 			} else if (strcmp(event_type, "bar_state_update") == 0) {
@@ -911,7 +925,6 @@ void ipc_client_handle_command(struct ipc_client *client, uint32_t payload_lengt
 
 exit_cleanup:
 	free(buf);
-	return;
 }
 
 bool ipc_send_reply(struct ipc_client *client, enum ipc_command_type payload_type,
@@ -955,7 +968,5 @@ bool ipc_send_reply(struct ipc_client *client, enum ipc_command_type payload_typ
 				ipc_client_handle_writable, client);
 	}
 
-	sway_log(SWAY_DEBUG, "Added IPC reply of type 0x%x to client %d queue: %s",
-		payload_type, client->fd, payload);
 	return true;
 }
