@@ -3,10 +3,11 @@
 #include <drm_fourcc.h>
 #include <stdint.h>
 #include <stdlib.h>
-#include <string.h>
-#include <strings.h>
 #include <wayland-server-core.h>
+#include <wlr/types/wlr_linux_dmabuf_v1.h>
 #include <wlr/types/wlr_output_layout.h>
+#include <wlr/types/wlr_subcompositor.h>
+#include "linux-dmabuf-unstable-v1-protocol.h"
 #include "cairo_util.h"
 #include "pango.h"
 #include "sway/config.h"
@@ -17,9 +18,11 @@
 #include "sway/ipc-server.h"
 #include "sway/output.h"
 #include "sway/server.h"
+#include "sway/surface.h"
 #include "sway/tree/arrange.h"
 #include "sway/tree/view.h"
 #include "sway/tree/workspace.h"
+#include "sway/xdg_decoration.h"
 #include "list.h"
 #include "log.h"
 #include "stringop.h"
@@ -43,7 +46,7 @@ struct sway_container *container_create(struct sway_view *view) {
 	c->outputs = create_list();
 
 	wl_signal_init(&c->events.destroy);
-	wl_signal_emit(&root->events.new_node, &c->node);
+	wl_signal_emit_mutable(&root->events.new_node, &c->node);
 
 	return c;
 }
@@ -63,6 +66,7 @@ void container_destroy(struct sway_container *con) {
 	wlr_texture_destroy(con->title_focused_inactive);
 	wlr_texture_destroy(con->title_unfocused);
 	wlr_texture_destroy(con->title_urgent);
+	wlr_texture_destroy(con->title_focused_tab_title);
 	list_free(con->pending.children);
 	list_free(con->current.children);
 	list_free(con->outputs);
@@ -72,11 +76,10 @@ void container_destroy(struct sway_container *con) {
 	wlr_texture_destroy(con->marks_focused_inactive);
 	wlr_texture_destroy(con->marks_unfocused);
 	wlr_texture_destroy(con->marks_urgent);
+	wlr_texture_destroy(con->marks_focused_tab_title);
 
-	if (con->view) {
-		if (con->view->container == con) {
-			con->view->container = NULL;
-		}
+	if (con->view && con->view->container == con) {
+		con->view->container = NULL;
 		if (con->view->destroying) {
 			view_destroy(con->view);
 		}
@@ -98,7 +101,7 @@ void container_begin_destroy(struct sway_container *con) {
 		container_fullscreen_disable(con);
 	}
 
-	wl_signal_emit(&con->node.events.destroy, &con->node);
+	wl_signal_emit_mutable(&con->node.events.destroy, &con->node);
 
 	container_end_mouse_operation(con);
 
@@ -188,7 +191,7 @@ static struct sway_container *surface_at_view(struct sway_container *con, double
 #endif
 	case SWAY_VIEW_XDG_SHELL:
 		_surface = wlr_xdg_surface_surface_at(
-				view->wlr_xdg_surface,
+				view->wlr_xdg_toplevel->base,
 				view_sx, view_sy, &_sx, &_sy);
 		break;
 	}
@@ -320,10 +323,10 @@ static struct sway_container *view_container_content_at(struct sway_node *parent
 
 	struct sway_container *container = parent->sway_container;
 	struct wlr_box box = {
-			.x = container->pending.content_x,
-			.y = container->pending.content_y,
-			.width = container->pending.content_width,
-			.height = container->pending.content_height,
+		.x = container->pending.content_x,
+		.y = container->pending.content_y,
+		.width = container->pending.content_width,
+		.height = container->pending.content_height,
 	};
 
 	if (wlr_box_contains_point(&box, lx, ly)) {
@@ -343,10 +346,10 @@ static struct sway_container *view_container_at(struct sway_node *parent,
 
 	struct sway_container *container = parent->sway_container;
 	struct wlr_box box = {
-			.x = container->pending.x,
-			.y = container->pending.y,
-			.width = container->pending.width,
-			.height = container->pending.height,
+		.x = container->pending.x,
+		.y = container->pending.y,
+		.width = container->pending.width,
+		.height = container->pending.height,
 	};
 
 	if (wlr_box_contains_point(&box, lx, ly)) {
@@ -381,19 +384,17 @@ struct sway_container *tiling_container_at(struct sway_node *parent,
 }
 
 static bool surface_is_popup(struct wlr_surface *surface) {
-	if (wlr_surface_is_xdg_surface(surface)) {
-		struct wlr_xdg_surface *xdg_surface =
-			wlr_xdg_surface_from_wlr_surface(surface);
-		while (xdg_surface && xdg_surface->role != WLR_XDG_SURFACE_ROLE_NONE) {
-			if (xdg_surface->role == WLR_XDG_SURFACE_ROLE_POPUP) {
-				return true;
-			}
-			xdg_surface = xdg_surface->toplevel->parent;
+	while (wlr_xdg_surface_try_from_wlr_surface(surface) == NULL) {
+		struct wlr_subsurface *subsurface =
+			wlr_subsurface_try_from_wlr_surface(surface);
+		if (subsurface == NULL) {
+			return false;
 		}
-		return false;
+		surface = subsurface->parent;
 	}
-
-	return false;
+	struct wlr_xdg_surface *xdg_surface =
+		wlr_xdg_surface_try_from_wlr_surface(surface);
+	return xdg_surface->role == WLR_XDG_SURFACE_ROLE_POPUP && xdg_surface->popup != NULL;
 }
 
 struct sway_container *container_at(struct sway_workspace *workspace,
@@ -492,6 +493,76 @@ struct sway_output *container_get_effective_output(struct sway_container *con) {
 	return con->outputs->items[con->outputs->length - 1];
 }
 
+static void render_titlebar_text_texture(struct sway_output *output,
+		struct sway_container *con, struct wlr_texture **texture,
+		struct border_colors *class, bool pango_markup, char *text) {
+	double scale = output->wlr_output->scale;
+	int width = 0;
+	int height = config->font_height * scale;
+	int baseline;
+
+	// We must use a non-nil cairo_t for cairo_set_font_options to work.
+	// Therefore, we cannot use cairo_create(NULL).
+	cairo_surface_t *dummy_surface = cairo_image_surface_create(
+			CAIRO_FORMAT_ARGB32, 0, 0);
+	cairo_t *c = cairo_create(dummy_surface);
+	cairo_set_antialias(c, CAIRO_ANTIALIAS_BEST);
+	cairo_font_options_t *fo = cairo_font_options_create();
+	if (output->wlr_output->subpixel == WL_OUTPUT_SUBPIXEL_NONE) {
+		cairo_font_options_set_antialias(fo, CAIRO_ANTIALIAS_GRAY);
+	} else {
+		cairo_font_options_set_antialias(fo, CAIRO_ANTIALIAS_SUBPIXEL);
+		cairo_font_options_set_subpixel_order(fo,
+			to_cairo_subpixel_order(output->wlr_output->subpixel));
+	}
+	cairo_set_font_options(c, fo);
+	get_text_size(c, config->font_description, &width, NULL, &baseline, scale,
+			config->pango_markup, "%s", text);
+	cairo_surface_destroy(dummy_surface);
+	cairo_destroy(c);
+
+	if (width == 0 || height == 0) {
+		return;
+	}
+
+	if (height > config->font_height * scale) {
+		height = config->font_height * scale;
+	}
+
+	cairo_surface_t *surface = cairo_image_surface_create(
+			CAIRO_FORMAT_ARGB32, width, height);
+	cairo_status_t status = cairo_surface_status(surface);
+	if (status != CAIRO_STATUS_SUCCESS) {
+		sway_log(SWAY_ERROR, "cairo_image_surface_create failed: %s",
+			cairo_status_to_string(status));
+		return;
+	}
+
+	cairo_t *cairo = cairo_create(surface);
+	cairo_set_antialias(cairo, CAIRO_ANTIALIAS_BEST);
+	cairo_set_font_options(cairo, fo);
+	cairo_font_options_destroy(fo);
+	cairo_set_source_rgba(cairo, class->background[0], class->background[1],
+			class->background[2], class->background[3]);
+	cairo_paint(cairo);
+	PangoContext *pango = pango_cairo_create_context(cairo);
+	cairo_set_source_rgba(cairo, class->text[0], class->text[1],
+			class->text[2], class->text[3]);
+	cairo_move_to(cairo, 0, config->font_baseline * scale - baseline);
+
+	render_text(cairo, config->font_description, scale, pango_markup, "%s", text);
+
+	cairo_surface_flush(surface);
+	unsigned char *data = cairo_image_surface_get_data(surface);
+	int stride = cairo_image_surface_get_stride(surface);
+	struct wlr_renderer *renderer = output->wlr_output->renderer;
+	*texture = wlr_texture_from_pixels(
+			renderer, DRM_FORMAT_ARGB8888, stride, width, height, data);
+	cairo_surface_destroy(surface);
+	g_object_unref(pango);
+	cairo_destroy(cairo);
+}
+
 static void update_title_texture(struct sway_container *con,
 		struct wlr_texture **texture, struct border_colors *class) {
 	struct sway_output *output = container_get_effective_output(con);
@@ -506,62 +577,8 @@ static void update_title_texture(struct sway_container *con,
 		return;
 	}
 
-	double scale = output->wlr_output->scale;
-	int width = 0;
-	int height = con->title_height * scale;
-
-	// We must use a non-nil cairo_t for cairo_set_font_options to work.
-	// Therefore, we cannot use cairo_create(NULL).
-	cairo_surface_t *dummy_surface = cairo_image_surface_create(
-			CAIRO_FORMAT_ARGB32, 0, 0);
-	cairo_t *c = cairo_create(dummy_surface);
-	cairo_set_antialias(c, CAIRO_ANTIALIAS_BEST);
-	cairo_font_options_t *fo = cairo_font_options_create();
-	cairo_font_options_set_hint_style(fo, CAIRO_HINT_STYLE_FULL);
-	if (output->wlr_output->subpixel == WL_OUTPUT_SUBPIXEL_NONE) {
-		cairo_font_options_set_antialias(fo, CAIRO_ANTIALIAS_GRAY);
-	} else {
-		cairo_font_options_set_antialias(fo, CAIRO_ANTIALIAS_SUBPIXEL);
-		cairo_font_options_set_subpixel_order(fo,
-			to_cairo_subpixel_order(output->wlr_output->subpixel));
-	}
-	cairo_set_font_options(c, fo);
-	get_text_size(c, config->font, &width, NULL, NULL, scale,
-			config->pango_markup, "%s", con->formatted_title);
-	cairo_surface_destroy(dummy_surface);
-	cairo_destroy(c);
-
-	if (width == 0 || height == 0) {
-		return;
-	}
-
-	cairo_surface_t *surface = cairo_image_surface_create(
-			CAIRO_FORMAT_ARGB32, width, height);
-	cairo_t *cairo = cairo_create(surface);
-	cairo_set_antialias(cairo, CAIRO_ANTIALIAS_BEST);
-	cairo_set_font_options(cairo, fo);
-	cairo_font_options_destroy(fo);
-	cairo_set_source_rgba(cairo, class->background[0], class->background[1],
-			class->background[2], class->background[3]);
-	cairo_paint(cairo);
-	PangoContext *pango = pango_cairo_create_context(cairo);
-	cairo_set_source_rgba(cairo, class->text[0], class->text[1],
-			class->text[2], class->text[3]);
-	cairo_move_to(cairo, 0, 0);
-
-	pango_printf(cairo, config->font, scale, config->pango_markup,
-			"%s", con->formatted_title);
-
-	cairo_surface_flush(surface);
-	unsigned char *data = cairo_image_surface_get_data(surface);
-	int stride = cairo_format_stride_for_width(CAIRO_FORMAT_ARGB32, width);
-	struct wlr_renderer *renderer = wlr_backend_get_renderer(
-			output->wlr_output->backend);
-	*texture = wlr_texture_from_pixels(
-			renderer, DRM_FORMAT_ARGB8888, stride, width, height, data);
-	cairo_surface_destroy(surface);
-	g_object_unref(pango);
-	cairo_destroy(cairo);
+	render_titlebar_text_texture(output, con, texture, class,
+		config->pango_markup, con->formatted_title);
 }
 
 void container_update_title_textures(struct sway_container *container) {
@@ -573,24 +590,11 @@ void container_update_title_textures(struct sway_container *container) {
 			&config->border_colors.unfocused);
 	update_title_texture(container, &container->title_urgent,
 			&config->border_colors.urgent);
+	update_title_texture(container, &container->title_focused_tab_title,
+			&config->border_colors.focused_tab_title);
 	container_damage_whole(container);
 }
 
-void container_calculate_title_height(struct sway_container *container) {
-	if (!container->formatted_title) {
-		container->title_height = 0;
-		return;
-	}
-	cairo_t *cairo = cairo_create(NULL);
-	int height;
-	int baseline;
-	get_text_size(cairo, config->font, NULL, &height, &baseline, 1,
-			config->pango_markup, "%s", container->formatted_title);
-	cairo_destroy(cairo);
-	container->title_height = height;
-	container->title_baseline = baseline;
-}
-
 /**
  * Calculate and return the length of the tree representation.
  * An example tree representation is: V[Terminal, Firefox]
@@ -656,7 +660,6 @@ void container_update_representation(struct sway_container *con) {
 		}
 		container_build_representation(con->pending.layout, con->pending.children,
 				con->formatted_title);
-		container_calculate_title_height(con);
 		container_update_title_textures(con);
 	}
 	if (con->pending.parent) {
@@ -688,12 +691,13 @@ void floating_calculate_constraints(int *min_width, int *max_width,
 		*min_height = config->floating_minimum_height;
 	}
 
-	struct wlr_box *box = wlr_output_layout_get_box(root->output_layout, NULL);
+	struct wlr_box box;
+	wlr_output_layout_get_box(root->output_layout, NULL, &box);
 
 	if (config->floating_maximum_width == -1) { // no maximum
 		*max_width = INT_MAX;
 	} else if (config->floating_maximum_width == 0) { // automatic
-		*max_width = box->width;
+		*max_width = box.width;
 	} else {
 		*max_width = config->floating_maximum_width;
 	}
@@ -701,13 +705,28 @@ void floating_calculate_constraints(int *min_width, int *max_width,
 	if (config->floating_maximum_height == -1) { // no maximum
 		*max_height = INT_MAX;
 	} else if (config->floating_maximum_height == 0) { // automatic
-		*max_height = box->height;
+		*max_height = box.height;
 	} else {
 		*max_height = config->floating_maximum_height;
 	}
 
 }
 
+void floating_fix_coordinates(struct sway_container *con, struct wlr_box *old, struct wlr_box *new) {
+	if (!old->width || !old->height) {
+		// Fall back to centering on the workspace.
+		container_floating_move_to_center(con);
+	} else {
+		int rel_x = con->pending.x - old->x + (con->pending.width / 2);
+		int rel_y = con->pending.y - old->y + (con->pending.height / 2);
+
+		con->pending.x = new->x + (double)(rel_x * new->width) / old->width - (con->pending.width / 2);
+		con->pending.y = new->y + (double)(rel_y * new->height) / old->height - (con->pending.height / 2);
+
+		sway_log(SWAY_DEBUG, "Transformed container %p to coords (%f, %f)", con, con->pending.x, con->pending.y);
+	}
+}
+
 static void floating_natural_resize(struct sway_container *con) {
 	int min_width, max_width, min_height, max_height;
 	floating_calculate_constraints(&min_width, &max_width,
@@ -733,9 +752,9 @@ void container_floating_resize_and_center(struct sway_container *con) {
 		return;
 	}
 
-	struct wlr_box *ob = wlr_output_layout_get_box(root->output_layout,
-			ws->output->wlr_output);
-	if (!ob) {
+	struct wlr_box ob;
+	wlr_output_layout_get_box(root->output_layout, ws->output->wlr_output, &ob);
+	if (wlr_box_empty(&ob)) {
 		// On NOOP output. Will be called again when moved to an output
 		con->pending.x = 0;
 		con->pending.y = 0;
@@ -747,8 +766,8 @@ void container_floating_resize_and_center(struct sway_container *con) {
 	floating_natural_resize(con);
 	if (!con->view) {
 		if (con->pending.width > ws->width || con->pending.height > ws->height) {
-			con->pending.x = ob->x + (ob->width - con->pending.width) / 2;
-			con->pending.y = ob->y + (ob->height - con->pending.height) / 2;
+			con->pending.x = ob.x + (ob.width - con->pending.width) / 2;
+			con->pending.y = ob.y + (ob.height - con->pending.height) / 2;
 		} else {
 			con->pending.x = ws->x + (ws->width - con->pending.width) / 2;
 			con->pending.y = ws->y + (ws->height - con->pending.height) / 2;
@@ -756,8 +775,8 @@ void container_floating_resize_and_center(struct sway_container *con) {
 	} else {
 		if (con->pending.content_width > ws->width
 				|| con->pending.content_height > ws->height) {
-			con->pending.content_x = ob->x + (ob->width - con->pending.content_width) / 2;
-			con->pending.content_y = ob->y + (ob->height - con->pending.content_height) / 2;
+			con->pending.content_x = ob.x + (ob.width - con->pending.content_width) / 2;
+			con->pending.content_y = ob.y + (ob.height - con->pending.content_height) / 2;
 		} else {
 			con->pending.content_x = ws->x + (ws->width - con->pending.content_width) / 2;
 			con->pending.content_y = ws->y + (ws->height - con->pending.content_height) / 2;
@@ -779,11 +798,11 @@ void container_floating_set_default_size(struct sway_container *con) {
 	int min_width, max_width, min_height, max_height;
 	floating_calculate_constraints(&min_width, &max_width,
 			&min_height, &max_height);
-	struct wlr_box *box = calloc(1, sizeof(struct wlr_box));
-	workspace_get_box(con->pending.workspace, box);
+	struct wlr_box box;
+	workspace_get_box(con->pending.workspace, &box);
 
-	double width = fmax(min_width, fmin(box->width * 0.5, max_width));
-	double height = fmax(min_height, fmin(box->height * 0.75, max_height));
+	double width = fmax(min_width, fmin(box.width * 0.5, max_width));
+	double height = fmax(min_height, fmin(box.height * 0.75, max_height));
 	if (!con->view) {
 		con->pending.width = width;
 		con->pending.height = height;
@@ -792,8 +811,6 @@ void container_floating_set_default_size(struct sway_container *con) {
 		con->pending.content_height = height;
 		container_set_geometry_from_content(con);
 	}
-
-	free(box);
 }
 
 
@@ -825,6 +842,8 @@ void container_set_floating(struct sway_container *container, bool enable) {
 
 	struct sway_seat *seat = input_manager_current_seat();
 	struct sway_workspace *workspace = container->pending.workspace;
+	struct sway_container *focus = seat_get_focused_container(seat);
+	bool set_focus = focus == container;
 
 	if (enable) {
 		struct sway_container *old_parent = container->pending.parent;
@@ -833,12 +852,22 @@ void container_set_floating(struct sway_container *container, bool enable) {
 		if (container->view) {
 			view_set_tiled(container->view, false);
 			if (container->view->using_csd) {
+				container->saved_border = container->pending.border;
 				container->pending.border = B_CSD;
+				if (container->view->xdg_decoration) {
+					struct sway_xdg_decoration *deco = container->view->xdg_decoration;
+					wlr_xdg_toplevel_decoration_v1_set_mode(deco->wlr_xdg_decoration,
+							WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_CLIENT_SIDE);
+				}
 			}
 		}
 		container_floating_set_default_size(container);
 		container_floating_resize_and_center(container);
 		if (old_parent) {
+			if (set_focus) {
+				seat_set_raw_focus(seat, &old_parent->node);
+				seat_set_raw_focus(seat, &container->node);
+			}
 			container_reap_empty(old_parent);
 		}
 	} else {
@@ -850,7 +879,11 @@ void container_set_floating(struct sway_container *container, bool enable) {
 		struct sway_container *reference =
 			seat_get_focus_inactive_tiling(seat, workspace);
 		if (reference) {
-			container_add_sibling(reference, container, 1);
+			if (reference->view) {
+				container_add_sibling(reference, container, 1);
+			} else {
+				container_add_child(reference, container);
+			}
 			container->pending.width = reference->pending.width;
 			container->pending.height = reference->pending.height;
 		} else {
@@ -863,6 +896,11 @@ void container_set_floating(struct sway_container *container, bool enable) {
 			view_set_tiled(container->view, true);
 			if (container->view->using_csd) {
 				container->pending.border = container->saved_border;
+				if (container->view->xdg_decoration) {
+					struct sway_xdg_decoration *deco = container->view->xdg_decoration;
+					wlr_xdg_toplevel_decoration_v1_set_mode(deco->wlr_xdg_decoration,
+							WLR_XDG_TOPLEVEL_DECORATION_V1_MODE_SERVER_SIDE);
+				}
 			}
 		}
 		container->width_fraction = 0;
@@ -884,7 +922,7 @@ void container_set_geometry_from_content(struct sway_container *con) {
 	size_t border_width = 0;
 	size_t top = 0;
 
-	if (con->pending.border != B_CSD) {
+	if (con->pending.border != B_CSD && !con->pending.fullscreen_mode) {
 		border_width = con->pending.border_thickness * (con->pending.border != B_NONE);
 		top = con->pending.border == B_NORMAL ?
 			container_titlebar_height() : border_width;
@@ -1002,6 +1040,13 @@ void container_floating_move_to(struct sway_container *con,
 		workspace_add_floating(new_workspace, con);
 		arrange_workspace(old_workspace);
 		arrange_workspace(new_workspace);
+		// If the moved container was a visible scratchpad container, then
+		// update its transform.
+		if (con->scratchpad) {
+			struct wlr_box output_box;
+			output_get_box(new_output, &output_box);
+			con->transform = output_box;
+		}
 		workspace_detect_urgent(old_workspace);
 		workspace_detect_urgent(new_workspace);
 	}
@@ -1033,18 +1078,44 @@ void container_end_mouse_operation(struct sway_container *container) {
 	}
 }
 
-static void set_fullscreen_iterator(struct sway_container *con, void *data) {
+static void set_fullscreen(struct sway_container *con, bool enable) {
 	if (!con->view) {
 		return;
 	}
 	if (con->view->impl->set_fullscreen) {
-		bool *enable = data;
-		con->view->impl->set_fullscreen(con->view, *enable);
+		con->view->impl->set_fullscreen(con->view, enable);
 		if (con->view->foreign_toplevel) {
 			wlr_foreign_toplevel_handle_v1_set_fullscreen(
-				con->view->foreign_toplevel, *enable);
+				con->view->foreign_toplevel, enable);
 		}
 	}
+
+	if (!server.linux_dmabuf_v1 || !con->view->surface) {
+		return;
+	}
+	if (!enable) {
+		wlr_linux_dmabuf_v1_set_surface_feedback(server.linux_dmabuf_v1,
+			con->view->surface, NULL);
+		return;
+	}
+
+	if (!con->pending.workspace || !con->pending.workspace->output) {
+		return;
+	}
+
+	struct sway_output *output = con->pending.workspace->output;
+
+	const struct wlr_linux_dmabuf_feedback_v1_init_options options = {
+		.main_renderer = server.renderer,
+		.scanout_primary_output = output->wlr_output,
+	};
+	struct wlr_linux_dmabuf_feedback_v1 feedback = {0};
+	if (!wlr_linux_dmabuf_feedback_v1_init_with_options(&feedback, &options)) {
+		return;
+	}
+	wlr_linux_dmabuf_v1_set_surface_feedback(server.linux_dmabuf_v1,
+		con->view->surface, &feedback);
+	wlr_linux_dmabuf_feedback_v1_finish(&feedback);
 }
 
 static void container_fullscreen_workspace(struct sway_container *con) {
@@ -1052,9 +1123,7 @@ static void container_fullscreen_workspace(struct sway_container *con) {
 				"Expected a non-fullscreen container")) {
 		return;
 	}
-	bool enable = true;
-	set_fullscreen_iterator(con, &enable);
-	container_for_each_child(con, set_fullscreen_iterator, &enable);
+	set_fullscreen(con, true);
 	con->pending.fullscreen_mode = FULLSCREEN_WORKSPACE;
 
 	con->saved_x = con->pending.x;
@@ -1088,9 +1157,7 @@ static void container_fullscreen_global(struct sway_container *con) {
 				"Expected a non-fullscreen container")) {
 		return;
 	}
-	bool enable = true;
-	set_fullscreen_iterator(con, &enable);
-	container_for_each_child(con, set_fullscreen_iterator, &enable);
+	set_fullscreen(con, true);
 
 	root->fullscreen_global = con;
 	con->saved_x = con->pending.x;
@@ -1116,9 +1183,7 @@ void container_fullscreen_disable(struct sway_container *con) {
 				"Expected a fullscreen container")) {
 		return;
 	}
-	bool enable = false;
-	set_fullscreen_iterator(con, &enable);
-	container_for_each_child(con, set_fullscreen_iterator, &enable);
+	set_fullscreen(con, false);
 
 	if (container_is_floating(con)) {
 		con->pending.x = con->saved_x;
@@ -1222,14 +1287,14 @@ bool container_is_fullscreen_or_child(struct sway_container *container) {
 
 static void surface_send_enter_iterator(struct wlr_surface *surface,
 		int x, int y, void *data) {
-	struct wlr_output *wlr_output = data;
-	wlr_surface_send_enter(surface, wlr_output);
+	struct sway_output *output = data;
+	surface_enter_output(surface, output);
 }
 
 static void surface_send_leave_iterator(struct wlr_surface *surface,
 		int x, int y, void *data) {
-	struct wlr_output *wlr_output = data;
-	wlr_surface_send_leave(surface, wlr_output);
+	struct sway_output *output = data;
+	surface_leave_output(surface, output);
 }
 
 void container_discover_outputs(struct sway_container *con) {
@@ -1255,7 +1320,7 @@ void container_discover_outputs(struct sway_container *con) {
 			sway_log(SWAY_DEBUG, "Container %p entered output %p", con, output);
 			if (con->view) {
 				view_for_each_surface(con->view,
-						surface_send_enter_iterator, output->wlr_output);
+						surface_send_enter_iterator, output);
 				if (con->view->foreign_toplevel) {
 					wlr_foreign_toplevel_handle_v1_output_enter(
 							con->view->foreign_toplevel, output->wlr_output);
@@ -1267,7 +1332,7 @@ void container_discover_outputs(struct sway_container *con) {
 			sway_log(SWAY_DEBUG, "Container %p left output %p", con, output);
 			if (con->view) {
 				view_for_each_surface(con->view,
-					surface_send_leave_iterator, output->wlr_output);
+					surface_send_leave_iterator, output);
 				if (con->view->foreign_toplevel) {
 					wlr_foreign_toplevel_handle_v1_output_leave(
 							con->view->foreign_toplevel, output->wlr_output);
@@ -1308,7 +1373,7 @@ list_t *container_get_siblings(struct sway_container *container) {
 	if (container->pending.parent) {
 		return container->pending.parent->pending.children;
 	}
-	if (container_is_scratchpad_hidden(container)) {
+	if (!container->pending.workspace) {
 		return NULL;
 	}
 	if (list_find(container->pending.workspace->tiling, container) != -1) {
@@ -1382,10 +1447,6 @@ void container_add_child(struct sway_container *parent,
 	child->pending.parent = parent;
 	child->pending.workspace = parent->pending.workspace;
 	container_for_each_child(child, set_workspace, NULL);
-	bool fullscreen = child->pending.fullscreen_mode != FULLSCREEN_NONE ||
-		parent->pending.fullscreen_mode != FULLSCREEN_NONE;
-	set_fullscreen_iterator(child, &fullscreen);
-	container_for_each_child(child, set_fullscreen_iterator, &fullscreen);
 	container_handle_fullscreen_reparent(child);
 	container_update_representation(parent);
 	node_set_dirty(&child->node);
@@ -1609,49 +1670,14 @@ static void update_marks_texture(struct sway_container *con,
 	for (int i = 0; i < con->marks->length; ++i) {
 		char *mark = con->marks->items[i];
 		if (mark[0] != '_') {
-			sprintf(part, "[%s]", mark);
+			snprintf(part, len + 1, "[%s]", mark);
 			strcat(buffer, part);
 		}
 	}
 	free(part);
 
-	double scale = output->wlr_output->scale;
-	int width = 0;
-	int height = con->title_height * scale;
+	render_titlebar_text_texture(output, con, texture, class, false, buffer);
 
-	cairo_t *c = cairo_create(NULL);
-	get_text_size(c, config->font, &width, NULL, NULL, scale, false,
-			"%s", buffer);
-	cairo_destroy(c);
-
-	if (width == 0 || height == 0) {
-		return;
-	}
-
-	cairo_surface_t *surface = cairo_image_surface_create(
-			CAIRO_FORMAT_ARGB32, width, height);
-	cairo_t *cairo = cairo_create(surface);
-	cairo_set_source_rgba(cairo, class->background[0], class->background[1],
-			class->background[2], class->background[3]);
-	cairo_paint(cairo);
-	PangoContext *pango = pango_cairo_create_context(cairo);
-	cairo_set_antialias(cairo, CAIRO_ANTIALIAS_BEST);
-	cairo_set_source_rgba(cairo, class->text[0], class->text[1],
-			class->text[2], class->text[3]);
-	cairo_move_to(cairo, 0, 0);
-
-	pango_printf(cairo, config->font, scale, false, "%s", buffer);
-
-	cairo_surface_flush(surface);
-	unsigned char *data = cairo_image_surface_get_data(surface);
-	int stride = cairo_format_stride_for_width(CAIRO_FORMAT_ARGB32, width);
-	struct wlr_renderer *renderer = wlr_backend_get_renderer(
-			output->wlr_output->backend);
-	*texture = wlr_texture_from_pixels(
-			renderer, DRM_FORMAT_ARGB8888, stride, width, height, data);
-	cairo_surface_destroy(surface);
-	g_object_unref(pango);
-	cairo_destroy(cairo);
 	free(buffer);
 }
 
@@ -1667,6 +1693,8 @@ void container_update_marks_textures(struct sway_container *con) {
 			&config->border_colors.unfocused);
 	update_marks_texture(con, &con->marks_urgent,
 			&config->border_colors.urgent);
+	update_marks_texture(con, &con->marks_focused_tab_title,
+			&config->border_colors.focused_tab_title);
 	container_damage_whole(con);
 }
 
@@ -1757,3 +1785,177 @@ int container_squash(struct sway_container *con) {
 	}
 	return change;
 }
+
+static void swap_places(struct sway_container *con1,
+		struct sway_container *con2) {
+	struct sway_container *temp = malloc(sizeof(struct sway_container));
+	temp->pending.x = con1->pending.x;
+	temp->pending.y = con1->pending.y;
+	temp->pending.width = con1->pending.width;
+	temp->pending.height = con1->pending.height;
+	temp->width_fraction = con1->width_fraction;
+	temp->height_fraction = con1->height_fraction;
+	temp->pending.parent = con1->pending.parent;
+	temp->pending.workspace = con1->pending.workspace;
+	bool temp_floating = container_is_floating(con1);
+
+	con1->pending.x = con2->pending.x;
+	con1->pending.y = con2->pending.y;
+	con1->pending.width = con2->pending.width;
+	con1->pending.height = con2->pending.height;
+	con1->width_fraction = con2->width_fraction;
+	con1->height_fraction = con2->height_fraction;
+
+	con2->pending.x = temp->pending.x;
+	con2->pending.y = temp->pending.y;
+	con2->pending.width = temp->pending.width;
+	con2->pending.height = temp->pending.height;
+	con2->width_fraction = temp->width_fraction;
+	con2->height_fraction = temp->height_fraction;
+
+	int temp_index = container_sibling_index(con1);
+	if (con2->pending.parent) {
+		container_insert_child(con2->pending.parent, con1,
+				container_sibling_index(con2));
+	} else if (container_is_floating(con2)) {
+		workspace_add_floating(con2->pending.workspace, con1);
+	} else {
+		workspace_insert_tiling(con2->pending.workspace, con1,
+				container_sibling_index(con2));
+	}
+	if (temp->pending.parent) {
+		container_insert_child(temp->pending.parent, con2, temp_index);
+	} else if (temp_floating) {
+		workspace_add_floating(temp->pending.workspace, con2);
+	} else {
+		workspace_insert_tiling(temp->pending.workspace, con2, temp_index);
+	}
+
+	free(temp);
+}
+
+static void swap_focus(struct sway_container *con1,
+		struct sway_container *con2, struct sway_seat *seat,
+		struct sway_container *focus) {
+	if (focus == con1 || focus == con2) {
+		struct sway_workspace *ws1 = con1->pending.workspace;
+		struct sway_workspace *ws2 = con2->pending.workspace;
+		enum sway_container_layout layout1 = container_parent_layout(con1);
+		enum sway_container_layout layout2 = container_parent_layout(con2);
+		if (focus == con1 && (layout2 == L_TABBED || layout2 == L_STACKED)) {
+			if (workspace_is_visible(ws2)) {
+				seat_set_focus(seat, &con2->node);
+			}
+			seat_set_focus_container(seat, ws1 != ws2 ? con2 : con1);
+		} else if (focus == con2 && (layout1 == L_TABBED
+					|| layout1 == L_STACKED)) {
+			if (workspace_is_visible(ws1)) {
+				seat_set_focus(seat, &con1->node);
+			}
+			seat_set_focus_container(seat, ws1 != ws2 ? con1 : con2);
+		} else if (ws1 != ws2) {
+			seat_set_focus_container(seat, focus == con1 ? con2 : con1);
+		} else {
+			seat_set_focus_container(seat, focus);
+		}
+	} else {
+		seat_set_focus_container(seat, focus);
+	}
+
+	if (root->fullscreen_global) {
+		seat_set_focus(seat,
+				seat_get_focus_inactive(seat, &root->fullscreen_global->node));
+	}
+}
+
+void container_swap(struct sway_container *con1, struct sway_container *con2) {
+	if (!sway_assert(con1 && con2, "Cannot swap with nothing")) {
+		return;
+	}
+	if (!sway_assert(!container_has_ancestor(con1, con2)
+				&& !container_has_ancestor(con2, con1),
+				"Cannot swap ancestor and descendant")) {
+		return;
+	}
+
+	sway_log(SWAY_DEBUG, "Swapping containers %zu and %zu",
+			con1->node.id, con2->node.id);
+
+	bool scratch1 = con1->scratchpad;
+	bool hidden1 = container_is_scratchpad_hidden(con1);
+	bool scratch2 = con2->scratchpad;
+	bool hidden2 = container_is_scratchpad_hidden(con2);
+	if (scratch1) {
+		if (hidden1) {
+			root_scratchpad_show(con1);
+		}
+		root_scratchpad_remove_container(con1);
+	}
+	if (scratch2) {
+		if (hidden2) {
+			root_scratchpad_show(con2);
+		}
+		root_scratchpad_remove_container(con2);
+	}
+
+	enum sway_fullscreen_mode fs1 = con1->pending.fullscreen_mode;
+	if (fs1) {
+		container_fullscreen_disable(con1);
+	}
+	enum sway_fullscreen_mode fs2 = con2->pending.fullscreen_mode;
+	if (fs2) {
+		container_fullscreen_disable(con2);
+	}
+
+	struct sway_seat *seat = input_manager_current_seat();
+	struct sway_container *focus = seat_get_focused_container(seat);
+	struct sway_workspace *vis1 =
+		output_get_active_workspace(con1->pending.workspace->output);
+	struct sway_workspace *vis2 =
+		output_get_active_workspace(con2->pending.workspace->output);
+	if (!sway_assert(vis1 && vis2, "con1 or con2 are on an output without a"
+				"workspace. This should not happen")) {
+		return;
+	}
+
+	char *stored_prev_name = NULL;
+	if (seat->prev_workspace_name) {
+		stored_prev_name = strdup(seat->prev_workspace_name);
+	}
+
+	swap_places(con1, con2);
+
+	if (!workspace_is_visible(vis1)) {
+		seat_set_focus(seat, seat_get_focus_inactive(seat, &vis1->node));
+	}
+	if (!workspace_is_visible(vis2)) {
+		seat_set_focus(seat, seat_get_focus_inactive(seat, &vis2->node));
+	}
+
+	swap_focus(con1, con2, seat, focus);
+
+	if (stored_prev_name) {
+		free(seat->prev_workspace_name);
+		seat->prev_workspace_name = stored_prev_name;
+	}
+
+	if (scratch1) {
+		root_scratchpad_add_container(con2, NULL);
+		if (!hidden1) {
+			root_scratchpad_show(con2);
+		}
+	}
+	if (scratch2) {
+		root_scratchpad_add_container(con1, NULL);
+		if (!hidden2) {
+			root_scratchpad_show(con1);
+		}
+	}
+
+	if (fs1) {
+		container_set_fullscreen(con2, fs1);
+	}
+	if (fs2) {
+		container_set_fullscreen(con1, fs2);
+	}
+}
