@@ -147,8 +147,10 @@ static bool ipc_parse_config(
 
 	json_object *font = json_object_object_get(bar_config, "font");
 	if (font) {
-		free(config->font);
-		config->font = parse_font(json_object_get_string(font));
+		pango_font_description_free(config->font_description);
+		char *font_value = parse_font(json_object_get_string(font));
+		config->font_description = pango_font_description_from_string(font_value);
+		free(font_value);
 	}
 
 	json_object *gaps = json_object_object_get(bar_config, "gaps");
@@ -424,12 +426,9 @@ bool ipc_initialize(struct swaybar *bar) {
 	}
 	free(res);
 
-	struct swaybar_config *config = bar->config;
-	char subscribe[128]; // suitably large buffer
-	len = snprintf(subscribe, 128,
-			"[ \"barconfig_update\" , \"bar_state_update\" %s %s ]",
-			config->binding_mode_indicator ? ", \"mode\"" : "",
-			config->workspace_buttons ? ", \"workspace\"" : "");
+	char *subscribe =
+		"[ \"barconfig_update\", \"bar_state_update\", \"mode\", \"workspace\" ]";
+	len = strlen(subscribe);
 	free(ipc_single_command(bar->ipc_event_socketfd,
 			IPC_SUBSCRIBE, subscribe, &len));
 	return true;
@@ -485,8 +484,7 @@ static bool handle_barconfig_update(struct swaybar *bar, const char *payload,
 			destroy_layer_surface(output);
 			wl_list_remove(&output->link);
 			wl_list_insert(&bar->unused_outputs, &output->link);
-		} else if (!oldcfg->font || !newcfg->font ||
-				strcmp(oldcfg->font, newcfg->font) != 0) {
+		} else if (!pango_font_description_equal(oldcfg->font_description, newcfg->font_description)) {
 			output->height = 0;  // force update height
 		}
 	}
